"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
from typing import Optional, Tuple

import skimage.io as sio
import skimage.morphology as sm
import skimage.measure as sme
import skimage.transform as st
import scipy.ndimage as ndimage

import keras
import keras.backend as kb
import keras.layers as kl
import keras.models as km
import numpy
import torch

from pyjamas.rimage.rimml.classifier_types import classifier_types
from pyjamas.rimage.rimutils import rimutils
from pyjamas.rimage.rimml.rimneuralnet import rimneuralnet


class ReSCUNet(rimneuralnet):
    CLASSIFIER_TYPE: str = classifier_types.RESCUNET.value
    CONCATENATION_LEVEL: int = 1

    def __init__(self, parameters: Optional[dict] = None):
        self.X_train_mask = None
        self.concatenation_level: int = parameters.get('concatenation_level', ReSCUNet.CONCATENATION_LEVEL)
        super().__init__(parameters)

        if type(self.step_sz) == int:  # backwards compatibility with how some models were saved
            self.step_sz = (self.step_sz, self.step_sz)

    def build_network(self, input_shape: Tuple, n_classes: int) -> km.Model:
        _epsilon = keras.ops.convert_to_tensor(kb.epsilon(), numpy.float32)

        # several input layers for data preprocessing steps
        input_layer = keras.Input(shape=input_shape, name="image_input")

        # the shape of the weight maps has to be such that it can be element-wise
        # multiplied to the softmax output.
        weight_input_layer = keras.Input(shape=input_shape[:2] + (n_classes,))
        mask_input_layer = keras.Input(shape=input_shape, name="mask_input")

        # ensure that the concatenation level is between 0 to 4, inclusive
        if self.concatenation_level < 0 or self.concatenation_level > 4:
            print("Invalid concatenation level, setting concatenation level to 0.")
            self.concatenation_level = 0

        curr_level = 0
        if self.concatenation_level == curr_level:
            in1 = [kl.Concatenate()([input_layer, mask_input_layer])]
        else:
            in1 = [input_layer]

        # adding the layers; image convolutions
        conv1 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(in1[0])
        conv1 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv1)
        conv1 = kl.Dropout(0.1)(conv1)
        mpool1 = kl.MaxPool2D()(conv1)

        if self.concatenation_level > curr_level:
            # mask convolutions
            convm1 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mask_input_layer)
            convm1 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(convm1)
            convm1 = kl.Dropout(0.1)(convm1)
            mpoolm1 = kl.MaxPool2D()(convm1)

        curr_level += 1
        if self.concatenation_level == curr_level:
            in2 = [kl.Concatenate()([mpool1, mpoolm1])]
        else:
            in2 = [mpool1]

        conv2 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(in2[0])
        conv2 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv2)
        conv2 = kl.Dropout(0.2)(conv2)
        mpool2 = kl.MaxPool2D()(conv2)

        if self.concatenation_level > curr_level:
            # mask convolutions
            convm2 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpoolm1)
            convm2 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(convm2)
            convm2 = kl.Dropout(0.2)(convm2)
            mpoolm2 = kl.MaxPool2D()(convm2)

        curr_level += 1
        if self.concatenation_level == curr_level:
            in3 = [kl.Concatenate()([mpool2, mpoolm2])]
        else:
            in3 = [mpool2]

        conv3 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(in3[0])
        conv3 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv3)
        conv3 = kl.Dropout(0.3)(conv3)
        mpool3 = kl.MaxPool2D()(conv3)

        if self.concatenation_level > curr_level:
            # mask convolutions
            convm3 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpoolm2)
            convm3 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(convm3)
            convm3 = kl.Dropout(0.3)(convm3)
            mpoolm3 = kl.MaxPool2D()(convm3)

        curr_level += 1
        if self.concatenation_level == curr_level:
            in4 = [kl.Concatenate()([mpool3, mpoolm3])]
        else:
            in4 = [mpool3]

        conv4 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(in4[0])
        conv4 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv4)
        conv4 = kl.Dropout(0.4)(conv4)
        mpool4 = kl.MaxPool2D()(conv4)

        if self.concatenation_level > curr_level:
            convm4 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpoolm3)
            convm4 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(convm4)
            convm4 = kl.Dropout(0.4)(convm4)
            mpoolm4 = kl.MaxPool2D()(convm4)

        curr_level += 1
        if self.concatenation_level == curr_level:
            in5 = [kl.Concatenate()([mpool4, mpoolm4])]
        else:
            in5 = [mpool4]

        conv5 = kl.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(in5[0])
        conv5 = kl.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv5)
        conv5 = kl.Dropout(0.5)(conv5)

        up6 = kl.Conv2DTranspose(512, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv5)
        conv6 = kl.Concatenate()([up6, conv4])
        conv6 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv6)
        conv6 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv6)
        conv6 = kl.Dropout(0.4)(conv6)

        up7 = kl.Conv2DTranspose(256, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv6)
        conv7 = kl.Concatenate()([up7, conv3])
        conv7 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv7)
        conv7 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv7)
        conv7 = kl.Dropout(0.3)(conv7)

        up8 = kl.Conv2DTranspose(128, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv7)
        conv8 = kl.Concatenate()([up8, conv2])
        conv8 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv8)
        conv8 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv8)
        conv8 = kl.Dropout(0.2)(conv8)

        up9 = kl.Conv2DTranspose(64, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv8)
        conv9 = kl.Concatenate()([up9, conv1])
        conv9 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = kl.Dropout(0.1)(conv9)

        c10 = kl.Conv2D(n_classes, 1, activation='softmax', kernel_initializer='he_normal', name="unet-activation")(
            conv9)

        # Add a few non-trainable layers to mimic the computation of the cross-entropy loss,
        # so that the actual loss function just has to perform the aggregation.
        c11 = kl.Lambda(lambda x: x / keras.ops.sum(x, len(x.shape) - 1, True))(c10)
        c11 = kl.Lambda(lambda x: keras.ops.clip(x, _epsilon, 1. - _epsilon))(c11)
        c11 = kl.Lambda(lambda x: keras.ops.log(x))(c11)
        weighted_sm = kl.multiply([c11, weight_input_layer])

        return km.Model(inputs=[input_layer, weight_input_layer, mask_input_layer], outputs=[weighted_sm])

    def _process_training_data(self) -> bool:
        """
        Loads and scales training data (previous masks).
        :return:
        """
        super()._process_training_data()

        train_ids = next(os.walk(self.positive_training_folder))[1]

        # Get and resize train previous masks
        self.X_train_mask = numpy.zeros((len(train_ids), self.train_image_size[0], self.train_image_size[1], 1), dtype=bool)

        for n, id_ in enumerate(train_ids):
            path = os.path.join(self.positive_training_folder, id_)

            prev_mask_file = os.path.join(path, "prev_mask", os.listdir(os.path.join(path, "prev_mask"))[0])
            prev_mask = sio.imread(prev_mask_file)
            prev_mask = numpy.expand_dims(st.resize(prev_mask, (self.train_image_size[0], self.train_image_size[1]), order=0,
                                                     mode='constant', preserve_range=True), axis=-1)

            max_value = numpy.max(numpy.max(prev_mask))
            prev_mask = numpy.round(numpy.divide(prev_mask, numpy.full(prev_mask.shape, max_value)))
            prev_mask = numpy.multiply(prev_mask, numpy.full(prev_mask.shape, max_value))
            prev_mask = prev_mask.clip(0)
            self.X_train_mask[n] = prev_mask

        return True

    def _get_training_inputs(self) -> list:
        return [self.X_train, self.W_train, self.X_train_mask]

    def predict(self, image: numpy.ndarray, prev_mask: numpy.ndarray) -> (numpy.ndarray, numpy.ndarray):
        if image is None or image is False:
            return False
        if prev_mask is None or prev_mask is False:
            return False

        if image.ndim == 3:
            image = image[0, :, :]
        if prev_mask.ndim == 3:
            prev_mask = prev_mask[0, :, :]

        testImage = image / self.scaler

        image_input = self.classifier.input[0]
        prev_mask_input = self.classifier.input[2]
        softmax_output = self.classifier.get_layer('unet-activation').output
        predictor = keras.Model(inputs=[image_input, prev_mask_input], outputs=softmax_output)

        with torch.no_grad():  # no gradient tracking (faster inference).
            # Keep these on GPU as PyTorch tensors for faster computation
            device = 'cpu'
            if torch.mps.is_available():
                device = torch.device("mps")
            elif torch.cuda.is_available():
                device = torch.device("cuda")

            testLabel = torch.zeros(testImage.shape, dtype=torch.bool, device=device)
            testProb = torch.zeros(testImage.shape, dtype=torch.float32, device=device)
            half_width = int(self.train_image_size[1] / 2)
            half_height = int(self.train_image_size[0] / 2)
            for animage, therow, thecol in rimutils.generate_subimages(testImage, self.train_image_size[0:2],
                                                                       self.step_sz, True):
                prev_mask_subimage = prev_mask[(therow - half_height):(therow + half_height),
                                               (thecol - half_width):(thecol + half_width)]

                if numpy.sum(prev_mask_subimage) == 0:  # empty prev mask in this subimage, skip area
                    continue

                yhat = predictor([numpy.expand_dims(animage, axis=0), numpy.expand_dims(prev_mask_subimage, axis=0)], training=False)
                yhat_class = torch.argmax(yhat[0], dim=-1)
                p = torch.amax(yhat[0], dim=-1)

                testLabel[(therow - half_height):(therow + half_height), (thecol - half_width):(thecol + half_width)] \
                    = torch.logical_or(testLabel[(therow - half_height):(therow + half_height),
                                       (thecol - half_width):(thecol + half_width)], yhat_class)
                testProb[(therow - half_height):(therow + half_height), (thecol - half_width):(thecol + half_width)] = p

            # Only convert to NumPy and leave the GPU at the very end
            testLabel = testLabel.cpu().numpy()
            testProb = testProb.cpu().numpy()

            # perform postprocessing to fill holes and take only the one object that best matches the previous mask
            testLabel = ndimage.binary_fill_holes(testLabel)
            labelled_mask = sme.label(testLabel)
            region_ids = numpy.unique(labelled_mask[labelled_mask > 0])
            if len(region_ids) > 1:
                best_IOU = None
                best_id = None
                for this_id in region_ids:
                    this_region = labelled_mask == this_id
                    intersection = numpy.sum(numpy.logical_and(this_region, prev_mask))
                    union = numpy.sum(numpy.logical_or(this_region, prev_mask))
                    this_IOU = intersection / union
                    if best_id is None or this_IOU > best_IOU:
                        best_IOU = this_IOU
                        best_id = this_id
                testLabel = labelled_mask == best_id

        if self.erosion_width is not None and self.erosion_width != 0:
            self.object_array = numpy.asarray(rimutils.extract_contours(
                sm.dilation(sm.label(sm.binary_erosion(testLabel, sm.square(self.erosion_width)), connectivity=1),
                            sm.square(self.erosion_width))), dtype=object)
        else:
            self.object_array = numpy.asarray(rimutils.extract_contours(sm.label(testLabel, connectivity=1)),
                                              dtype=object)
        self.prob_array = testProb

        return self.object_array.copy(), self.prob_array.copy()

    def save(self, filename: str) -> bool:
        self.save_dict.update({
            'classifier_type': self.CLASSIFIER_TYPE,
            'concatenation_level': self.concatenation_level,
        })
        return super().save(filename)



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional, Tuple

import skimage.morphology as sm
import keras
import keras.backend as kb
import keras.layers as kl
import keras.models as km
import numpy
import torch

from pyjamas.rimage.rimutils import rimutils
from pyjamas.rimage.rimml.rimneuralnet import rimneuralnet
from pyjamas.rimage.rimml.classifier_types import classifier_types


class UNet(rimneuralnet):

    CLASSIFIER_TYPE: str = classifier_types.UNET.value

    def __init__(self, parameters: Optional[dict] = None):
        super().__init__(parameters)

    def build_network(self, input_shape: Tuple, n_classes: int) -> km.Model:
        _epsilon = keras.ops.convert_to_tensor(kb.epsilon(), numpy.float32)

        # two inputs, one for the image and one for the weight maps
        ip = keras.Input(shape=input_shape, name="image_input")
        # the shape of the weight maps has to be such that it can be element-wise
        # multiplied to the softmax output.
        weight_ip = keras.Input(shape=input_shape[:2] + (n_classes,))

        # adding the layers
        conv1 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(ip)
        conv1 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv1)
        conv1 = kl.Dropout(0.1)(conv1)
        mpool1 = kl.MaxPool2D()(conv1)

        conv2 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpool1)
        conv2 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv2)
        conv2 = kl.Dropout(0.2)(conv2)
        mpool2 = kl.MaxPool2D()(conv2)

        conv3 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpool2)
        conv3 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv3)
        conv3 = kl.Dropout(0.3)(conv3)
        mpool3 = kl.MaxPool2D()(conv3)

        conv4 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpool3)
        conv4 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv4)
        conv4 = kl.Dropout(0.4)(conv4)
        mpool4 = kl.MaxPool2D()(conv4)

        conv5 = kl.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpool4)
        conv5 = kl.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv5)
        conv5 = kl.Dropout(0.5)(conv5)

        up6 = kl.Conv2DTranspose(512, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv5)
        conv6 = kl.Concatenate()([up6, conv4])
        conv6 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv6)
        conv6 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv6)
        conv6 = kl.Dropout(0.4)(conv6)

        up7 = kl.Conv2DTranspose(256, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv6)
        conv7 = kl.Concatenate()([up7, conv3])
        conv7 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv7)
        conv7 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv7)
        conv7 = kl.Dropout(0.3)(conv7)

        up8 = kl.Conv2DTranspose(128, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv7)
        conv8 = kl.Concatenate()([up8, conv2])
        conv8 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv8)
        conv8 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv8)
        conv8 = kl.Dropout(0.2)(conv8)

        up9 = kl.Conv2DTranspose(64, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv8)
        conv9 = kl.Concatenate()([up9, conv1])
        conv9 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = kl.Dropout(0.1)(conv9)

        c10 = kl.Conv2D(n_classes, 1, activation='softmax', kernel_initializer='he_normal', name="unet-activation")(
            conv9)

        # Add a few non-trainable layers to mimic the computation of the cross-entropy
        # loss, so that the actual loss function just has to perform the
        # aggregation.
        c11 = kl.Lambda(lambda x: x / keras.ops.sum(x, len(x.shape) - 1, True))(c10)
        c11 = kl.Lambda(lambda x: keras.ops.clip(x, _epsilon, 1. - _epsilon))(c11)
        c11 = kl.Lambda(lambda x: keras.ops.log(x))(c11)
        weighted_sm = kl.multiply([c11, weight_ip])

        return km.Model(inputs=[ip, weight_ip], outputs=[weighted_sm])

    def predict(self, image: numpy.ndarray) -> (numpy.ndarray, numpy.ndarray):
        if image is None or image is False:
            return False

        if image.ndim == 3:
            image = image[0, :, :]

        testImage = image / self.scaler

        image_input = self.classifier.input[0]
        softmax_output = self.classifier.get_layer('unet-activation').output
        predictor = keras.Model(inputs=image_input, outputs=softmax_output)

        with torch.no_grad():  # no gradient tracking (faster inference).
            device = 'cpu'
            if torch.mps.is_available():
                device = torch.device("mps")
            elif torch.cuda.is_available():
                device = torch.device("cuda")

            testLabel = torch.zeros(testImage.shape, dtype=torch.bool, device=device)
            testProb = torch.zeros(testImage.shape, dtype=torch.float32, device=device)

            half_width = int(self.train_image_size[1] / 2)
            half_height = int(self.train_image_size[0] / 2)

            for animage, therow, thecol in rimutils.generate_subimages(testImage, self.train_image_size[0:2],
                                                                       self.step_sz, True):
                yhat = predictor(numpy.expand_dims(animage, axis=0), training=False)

                yhat_class = torch.argmax(yhat[0], dim=-1)
                p = torch.amax(yhat[0], dim=-1)

                testLabel[(therow - half_height):(therow + half_height),
                (thecol - half_width):(thecol + half_width)] = torch.logical_or(
                    testLabel[(therow - half_height):(therow + half_height), (thecol - half_width):(thecol + half_width)],
                    yhat_class)
                testProb[(therow - half_height):(therow + half_height),
                (thecol - half_width):(thecol + half_width)] = p

            # Only convert to NumPy and leave the GPU at the very end
            testLabel = testLabel.cpu().numpy()
            testProb = testProb.cpu().numpy()

        if (self.erosion_width is not None) and (self.erosion_width != 0):
            self.object_array = numpy.asarray(rimutils.extract_contours(
                sm.dilation(sm.label(sm.binary_erosion(testLabel, sm.footprint_rectangle((self.erosion_width, self.erosion_width))), connectivity=2),
                            sm.footprint_rectangle((self.erosion_width, self.erosion_width)))), dtype=object)
        else:
            self.object_array = numpy.asarray(rimutils.extract_contours(sm.label(testLabel, connectivity=2)),
                                              dtype=object)
        self.prob_array = testProb

        return self.object_array.copy(), self.prob_array.copy()

    def save(self, filename: str) -> bool:
        self.save_dict.update({
            'classifier_type': self.CLASSIFIER_TYPE
        })
        return super().save(filename)


"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import re
from typing import Callable, Optional

import matplotlib.pyplot as plt
import matplotlib as mpl
import skimage.morphology as skm
from PyQt6 import QtWidgets, QtGui, QtCore

from pyjamas.pjscore import undo_modes
from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.rimage.rimcore import rimage
from pyjamas.dialogs.configurekeybindingsdialog import ConfigureKeyBindingsDialog
from pyjamas.dialogs.footprintdialog import FootprintDialog


class RCBOptions(RCallback):
    def cbUndo(self) -> bool:
        """
        Undo the most recent action: individual fiducials added or deleted; polylines added, deleted or moved; or multiple fiducials or polylines added or deleted simultaneously (e.g. after finding seeds, segmenting cells, or deleting all annotations). Also, image projections, inversion, smoothing, gradient, registration, crop or kymograph.

        :return: True if undo finished correctly, False if there was an error.
        """

        undoInfo = self.pjs.undo_stack.pop()
        if undoInfo is None:
            return False

        changetype = undoInfo.get("changetype")
        frame = undoInfo.get("frame")
        index = undoInfo.get("index")
        details = undoInfo.get("details")

        try:
            if changetype == undo_modes.FIDUCIAL_ADDED:
                self.pjs.fiducials[frame].pop(index)
            elif changetype == undo_modes.FIDUCIAL_DELETED:
                self.pjs.fiducials[frame].insert(index, details)
            elif changetype == undo_modes.POLYLINE_ADDED:
                self.pjs.polylines[frame].pop(index)
                self.pjs.polyline_ids[frame].pop(index)
            elif changetype == undo_modes.POLYLINE_DELETED:
                self.pjs.polylines[frame].append(details)
                self.pjs.polyline_ids[frame].append(index)
            elif changetype == undo_modes.POLYLINE_REPLACED:
                ind_list = [anind for anind in range(len(self.pjs.polyline_ids[frame]))
                            if self.pjs.polyline_ids[frame][anind] == index]
                self.pjs.polylines[frame][ind_list[0]] = details
            elif changetype == undo_modes.MULTI:
                if index[0]:
                    self.pjs.fiducials[frame[0]:frame[1]] = details[0]
                if index[1]:
                    self.pjs.polylines[frame[0]:frame[1]] = details[1]
                if index[2]:
                    self.pjs.polyline_ids[frame[0]:frame[1]] = details[2]
            elif changetype == undo_modes.IMAGE:
                self.pjs.io.cbLoadArray(frame)
                self.pjs.fiducials = details[0]
                self.pjs.polylines = details[1]
                self.pjs.polyline_ids = details[2]
                self.pjs.image.cbGoTo(index)

            if changetype == undo_modes.IMAGE or changetype == undo_modes.MULTI or frame == self.pjs.curslice:
                self.pjs.repaint()

        except Exception as an_exception:
            print(an_exception)
            self.pjs.undo_stack.clear()
            self.pjs.statusbar.showMessage("Something went wrong with the undo function. Undo stack cleared.")

            return False

        return True

    def cbSetBrushSize(self, sz: int=None) -> bool:
        """
        Set the size of the brush used to paint polygons.

        :param sz: brush size; if the value is None a dialog appears.
        :return: True if the brush size was changed, False otherwise.
        """

        brush_size: int = 0
        ok_flag: bool = None

        if sz not in [None, False]:
            brush_size = sz
            ok_flag = True
        else:
            # Read user input for brush size.
            brush_size, ok_flag = QtWidgets.QInputDialog.getInt(None, 'Set brush size: ', 'Enter new size: ',
                    self.pjs.brush_size, 1)

        if ok_flag and brush_size > 0:
            self.pjs.brush_size = brush_size
            self.pjs.repaint()
            return True

        else:
            return False

    def cbDisplayFiducialIDs(self) -> bool:
        """
        Toggle fiducial ids on/off.

        :return: True.
        """
        if self.pjs.display_fiducial_ids:
            self.pjs.display_fiducial_ids = False
        else:
            self.pjs.display_fiducial_ids = True
        self.pjs.repaint()

        return True

    def cbFramesPerSec(self, fps: Optional[int] = None) -> bool:
        """
        Set the number of frames per second to use when playing through the slices.

        :param fps: number of frames per second; if the value is None, a dialog appears.
        :return: True if the number of frames per second was changed, False otherwise.
        """

        thefps: int = 0
        ok_flag: bool = None

        if fps is not None and fps is not False:
            thefps = fps
            ok_flag = True
        else:
            # Read user input for fps.
            thefps, ok_flag = QtWidgets.QInputDialog.getInt(None, 'Set frames per second: ',
                                                            'Enter frames per second: ',
                                                            self.pjs.fps, 1)

        if ok_flag and thefps > 0:
            self.pjs.fps = thefps
            return True

        else:
            return False

    def cbSetCWD(self, folder_name: str = '') -> bool:
        """
        Set the current working directory.

        :param folder_name: absolute path to the current working directory; if the string is empty ('') or None, a dialog appears.
        :return: True if the directory was set (the selected directory must exist), False otherwise.
        """
        if folder_name == '' or folder_name is False or folder_name is None: # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            folder_name = QtWidgets.QFileDialog.getExistingDirectory(None, 'Set working folder to ...', self.pjs.cwd)

        # If cancel ...
        if folder_name == '':
            return False

        if os.path.exists(folder_name):
            self.pjs.cwd = os.path.abspath(folder_name)
            self.pjs.statusbar.showMessage(f"Working folder set to {self.pjs.cwd}.")
            return True

        else:
            return False

    def cbSetMarginSize(self, margin_size: Optional[int] = None) -> bool:
        """
        Set the size of margins used for cropping.

        :param margin_size: size of margins in pixels.
        :return: True if margin size was changed, False otherwise.
        """
        themargin: int = 0
        ok_flag: bool = True

        if margin_size is None or not margin_size:
            # Read user input for margin size.
            themargin, ok_flag = QtWidgets.QInputDialog.getInt(None, 'Set margin size: ',
                                                               'Enter new size: ', self.pjs.margin_size, 0,
                                                               min(self.pjs.height, self.pjs.width))
        else:
            themargin = margin_size
            ok_flag = True

        if ok_flag and (themargin >= 0):
            self.pjs.margin_size = themargin
            return True
        else:
            return False

    def cbSetImageProcessingCropSize(self, crop_size: Optional[int] = None) -> bool:
        """
        Set the size of the image crop that will be used for expensive computational operations (e.g. balloons).

        :param crop_size: size of the cropped image.
        :return: True if crop size was changed, False otherwise.
        """
        thesize: int = 0
        ok_flag: bool = True

        if crop_size is None or not crop_size:
            # Read user input for margin size.
            thesize, ok_flag = QtWidgets.QInputDialog.getInt(None, 'Set crop size: ',
                                                             'Enter new size: ', self.pjs.balloon_crop_size * 2, 0,
                                                             min(self.pjs.height, self.pjs.width))
        else:
            thesize = crop_size
            ok_flag = True

        if ok_flag and (thesize >= 0):
            self.pjs.balloon_crop_size = int(thesize / 2)
            return True
        else:
            return False

    def cbCropTracked(self) -> bool:
        """
        Toggles cropping with tracked polylines on or off.

        :return: True.
        """

        self.pjs.crop_tracked_polyline = not self.pjs.crop_tracked_polyline

        if self.pjs.crop_tracked_polyline:
            self.pjs.statusbar.showMessage('Crop tracked polyline turned ON.')
        else:
            self.pjs.statusbar.showMessage('Crop tracked polyline turned OFF.')

        return True

    def cbChangeDisplayTheme(self) -> bool:
        """
        Toggles display between light theme and dark theme.

        :return: True.
        """
        
        if self.pjs.darktheme is False:
            self.pjs.app.setStyleSheet("""
                QMainWindow{
                    background-color: #242424;
                }
                QMenuBar {
                    background-color: #242424;
                    color: white;
                }
                QMenuBar::item::selected {
                    background-color: #6e6e6e;
                    color: white;
                }
                QMenu {
                    background-color: #242424;
                    color: white;
                }
                QMenu::item::selected{
                    background-color: #6e6e6e;
                    color: white;
                }
                QWidget{
                    background: #242424;
                    color: white;
                }
                QPushButton{
                    background-color: #6e6e6e;
                }
                QPushButton:pressed{
                    background-color: #575757;
                }
                QLabel{
                    background-color: transparent;
                }
            """)
            plt.style.use('dark_background')
            mpl.rc('figure', facecolor = '#242424', edgecolor = 'white')
            mpl.rc('axes', facecolor = '#242424', edgecolor = 'white')
            self.pjs.darktheme = True

        else:
            self.pjs.app.setStyleSheet("")
            plt.style.use('default')
            mpl.rcdefaults()
            self.pjs.darktheme = False
        
        return True

    def cbCloseAllPolylines(self):
        """
        Toggles closing all polylines loaded from file ON/OFF.

        :return: True.
        """

        self.pjs.close_all_polylines = not self.pjs.close_all_polylines

        if self.pjs.close_all_polylines:
            self.pjs.statusbar.showMessage('Close all polylines turned ON.')
        else:
            self.pjs.statusbar.showMessage('Close all polylines turned OFF.')

        return True

    def cbSetLiveWireSmoothGaussian(self, sigma: Optional[float] = None) -> bool:
        """
        Set the standard deviation of the Gaussian used to smoothen images for LiveWire segmentation.
        A weight of 0.0 is no smoothing.

        :param sigma: standard deviation of the Gaussian used for image smoothing.
        :return: True if sigma was changed, False otherwise.
        """
        thesigma: float = 0.
        ok_flag: bool = True

        if sigma is None or not sigma:
            # Read user input for margin size.
            thesigma, ok_flag = QtWidgets.QInputDialog.getDouble(None, 'Set Gaussian sigma for LiveWire segmentation: ',
                                                                  'Enter new sigma: ',
                                                                 self.pjs.livewire_gaussian_sigma, 0)
        else:
            thesigma = sigma
            ok_flag = True

        if ok_flag and (thesigma >= 0.):
            self.pjs.livewire_gaussian_sigma = thesigma
            return True
        else:
            return False

    def cbSetLivewireShortestPathFunction(self, shortest_path_fn: Optional[str]=None):
        """
        Set the function used to calculate the shortest path between two pixels using LiveWire segmentation.

        :param shortest_path_fn: name of the function to calculate the shortest path between two pixels (see rimage.rimcore.rimage.livewire_shortest_path_fns for examples).
        :return: True if function was changed, False otherwise.
        """
        thefn = None
        if shortest_path_fn is None or not shortest_path_fn:
            fn_names = tuple(rimage.livewire_shortest_path_fns.keys())
            thefn, ok_flag = QtWidgets.QInputDialog.getItem(None, 'Select shortest path function for LiveWire segmentation: ',
                                            'Function', fn_names, fn_names.index(self.pjs.livewire_shortest_path_fn), False)

        else:
            thefn = shortest_path_fn
            ok_flag = True

        if ok_flag:
            self.pjs.livewire_shortest_path_fn = thefn
            return True
        else:
            return False

    def cbChangeKeyBindings(self, key_bindings: Optional[dict] = None):
        """
        Configure which keyboard shortcuts are used for which PyJAMAS functions.

        :param key_bindings: dict of dicts of format {'menu1_title': {'action1_text': 'action1_shortcut'}}
        :return: True if key bindings were changed, False otherwise.
        """
        continue_flag: bool = True
        if key_bindings is None or key_bindings is False:
            key_bindings = self.pjs.get_key_bindings()
            dialog = QtWidgets.QDialog()
            ui = ConfigureKeyBindingsDialog()
            ui.setupUi(dialog, key_bindings)

            dialog.exec()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted

            if continue_flag:
                key_bindings = ui.parameters()
            dialog.close()

        if continue_flag:
            return self.configure_key_bindings(key_bindings)
        else:
            return False

    def configure_key_bindings(self, key_bindings: dict):
        pjs_menus = self.pjs.menubar.children()[1:]  # skip zeroth index (QToolBar)
        for menu_index, this_menu in enumerate(key_bindings.keys()):
            these_actions = self.pjs.get_menu_children(pjs_menus[menu_index])
            for option_index, option_shortcut in enumerate(key_bindings[this_menu].values()):
                old_shortcut: QtGui.QShortcut = these_actions[option_index].data()['shortcut']
                if old_shortcut.key().toString() != option_shortcut:
                    these_actions[option_index].setShortcut(option_shortcut)
                    old_shortcut.setKey(QtGui.QKeySequence(option_shortcut))
        return True

    def cbToggleHighlightOnHover(self):
        """quit()

        Toggle highlighting objects on hover on/off.
        :return: True.
        """
        self.pjs.highlight_on_hover = not self.pjs.highlight_on_hover
        self.pjs.repaint()

        return True

    def cbChangeFootprint(self, footprint: str):
        """
        Change the footprint for dilations and erosions
        :return: True.
        """
        if type(footprint) is not str or not getattr(skm, footprint, False):
            dialog = QtWidgets.QDialog()
            ui = FootprintDialog()
            ui.setupUi(dialog, footprint=self.pjs.footprint)
            dialog.exec()
            dialog.show()
            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
            footprint = ui.parameters()
            dialog.close()
            if not continue_flag:
                return False

        self.pjs.footprint = footprint
        return True



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import skimage.morphology as sm
import skimage.measure as sme
import scipy.ndimage as ndimage
import scipy.stats as stats

import keras
import numpy
import torch

from pyjamas.rimage.rimml.classifier_types import classifier_types
from pyjamas.rimage.rimutils import rimutils

from pyjamas.rimage.rimml.rimrescunet import ReSCUNet


class QuickieNet(ReSCUNet):
    CLASSIFIER_TYPE = classifier_types.QUICKIENET.value
    def predict(self, image: numpy.ndarray, prev_mask: numpy.ndarray) -> (numpy.ndarray, numpy.ndarray):
        if image is None or image is False:
            return False
        if prev_mask is None or prev_mask is False:
            return False

        if image.ndim == 3:
            image = image[0, :, :]
        if prev_mask.ndim == 3:
            prev_mask = prev_mask[0, :, :]

        if prev_mask.dtype == bool:
            prev_labels = sme.label(prev_mask, connectivity=2)
        else:
            prev_labels = prev_mask.copy()
            prev_mask = prev_mask > 0
        prev_ids = numpy.unique(prev_labels[prev_mask])

        testImage = image / self.scaler

        image_input = self.classifier.input[0]
        prev_mask_input = self.classifier.input[2]
        softmax_output = self.classifier.get_layer('unet-activation').output
        predictor = keras.Model(inputs=[image_input, prev_mask_input], outputs=softmax_output)

        with torch.no_grad():  # no gradient tracking (faster inference).
            # Keep these on GPU as PyTorch tensors for faster computation
            device = 'cpu'
            if torch.mps.is_available():
                device = torch.device("mps")
            elif torch.cuda.is_available():
                device = torch.device("cuda")

            testLabel = torch.zeros(testImage.shape, dtype=torch.bool, device=device)
            testProb = torch.zeros(testImage.shape, dtype=torch.float32, device=device)

            half_width = int(self.train_image_size[1] / 2)
            half_height = int(self.train_image_size[0] / 2)

            for animage, therow, thecol in rimutils.generate_subimages(testImage, self.train_image_size[0:2],
                                                                       self.step_sz, True):
                prev_mask_subimage = prev_mask[(therow - half_height):(therow + half_height),
                                               (thecol - half_width):(thecol + half_width)]

                if numpy.sum(prev_mask_subimage) == 0:  # empty prev mask in this subimage, skip area
                    continue

                yhat = predictor([numpy.expand_dims(animage, axis=0), numpy.expand_dims(prev_mask_subimage, axis=0)], training=False)
                yhat_class = torch.argmax(yhat[0], dim=-1)
                p = torch.amax(yhat[0], dim=-1)

                testLabel[(therow - half_height):(therow + half_height), (thecol - half_width):(thecol + half_width)] \
                    = torch.logical_or(testLabel[(therow - half_height):(therow + half_height),
                                       (thecol - half_width):(thecol + half_width)], yhat_class)
                testProb[(therow - half_height):(therow + half_height), (thecol - half_width):(thecol + half_width)] = p

            # Only convert to NumPy and leave the GPU at the very end
            testLabel = testLabel.cpu().numpy()
            testProb = testProb.cpu().numpy()

        testLabel = ndimage.binary_fill_holes(testLabel)
        labelled_mask = sme.label(testLabel, connectivity=2)
        region_ids = numpy.unique(labelled_mask[labelled_mask > 0])

        if len(region_ids) > 1:
            self.object_array = numpy.empty(prev_ids.shape, dtype=object)

            for index, this_id in enumerate(prev_ids):
                this_region_before = prev_labels == this_id
                # Find the mode of the prev_labels images masked with the current label object, this_region
                best_id = stats.mode(labelled_mask[this_region_before]).mode

                testLabel_singleobject = numpy.zeros(testLabel.shape, dtype=numpy.bool)
                testLabel_singleobject[labelled_mask==best_id] = True

                thecontour = []

                if self.erosion_width is not None and self.erosion_width != 0:
                    thecontour = rimutils.extract_contours(
                        sm.dilation(sm.label(sm.binary_erosion(testLabel_singleobject, sm.footprint_rectangle((self.erosion_width, self.erosion_width))), connectivity=2),
                                    sm.footprint_rectangle((self.erosion_width, self.erosion_width))), fully_connected='high')
                else:
                    thecontour = rimutils.extract_contours(testLabel_singleobject, fully_connected='high')

                if len(thecontour) == 1:  # switch this to > 0 to allow returning more than one contour.
                    self.object_array[index] = numpy.squeeze(numpy.asarray(thecontour, dtype=object))
                else:
                    print(f"Error extracting a contour for polyline id {this_id}. Try re-prompting or smoothing the contour?")
                    return numpy.empty(0), numpy.empty(0)
        self.prob_array = testProb

        return self.object_array, self.prob_array



from pyjamas.rplugins.base import PJSPluginABC

class PJSPlugin(PJSPluginABC):
    def name(self) -> str:
        return "Sample plugin"

    def run(self, parameters: dict) -> bool:
        print("Hello my friends!")
        self.pjs.statusbar.showMessage("Hello my friends!")

        return True


"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PyQt6 import QtCore, QtWidgets


class DropDownDialog(object):
    text = ''
    index = None
    accepted = False

    def __init__(self, items: list = [], title: str = ''):
        super().__init__()
        self.dialog = QtWidgets.QDialog()
        self.setupUi(self.dialog)
        self.comboBox.addItems(items)
        self.dialog.setWindowTitle(title)

    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(200, 100)

        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(10, 50, 160, 30))
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(
            QtWidgets.QDialogButtonBox.StandardButton.Cancel | QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")

        self.comboBox = QtWidgets.QComboBox(Dialog)
        self.comboBox.setObjectName("comboBox")
        self.comboBox.setGeometry(QtCore.QRect(10, 10, 180, 30))

        self.retranslateUi()
        self.buttonBox.accepted.connect(self.ok_button)
        self.buttonBox.rejected.connect(self.cancel_button)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self):
        _translate = QtCore.QCoreApplication.translate

    def show(self):
        self.dialog.exec()

    def ok_button(self) -> bool:
        self.dialog.close()
        self.accepted = True
        return True

    def cancel_button(self) -> bool:
        self.dialog.close()
        self.accepted = False
        return False

    def result(self) -> bool:
        return self.accepted

    def parameters(self) -> str:
        self.text = self.comboBox.currentText()
        self.index = self.comboBox.currentIndex()

        return {'text': self.text,
                'index': self.index}



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional

import numpy
from PyQt6 import QtGui, QtCore

from pyjamas.pjsthreads import ThreadSignals
from pyjamas.rimage.rimml.batchml import BatchML
from pyjamas.rimage.rimml.rimneuralnet import rimneuralnet
from pyjamas.rimage.rimutils import rimutils


class BatchRecurrentNeuralNet(BatchML):
    def __init__(self, n_frames: int, classifier: Optional[rimneuralnet] = None):
        super().__init__(n_frames, classifier)
        self.prob_arrays = [numpy.empty((1, 0)) for i in range(self.n_frames)]
        self.object_arrays = [numpy.empty((1, 0)) for i in range(self.n_frames)]
        self.object_ids = [numpy.empty((1, 0)) for i in range(self.n_frames)]

    def fit(self, stop_signal: Optional[ThreadSignals] = None) -> bool:
        if self.image_classifier is None:
            return False

        if stop_signal is not None:
            stop_signal.emit('Launching classifier training ...')

        self.image_classifier.fit()

        if stop_signal is not None:
            stop_signal.emit('Classifier training completed.')
        return True

    def predict(self, slices: numpy.ndarray, indices: numpy.ndarray, progress_signal: Optional[ThreadSignals] = None,
                annotations: Optional[numpy.ndarray] = None, ids: Optional[list] = None) -> bool:

        # Make sure that the slices are in a 1D numpy array.
        indices = numpy.atleast_1d(indices)
        num_slices = len(indices)

        # Decide whether to go forward or backwards.
        apply_backwards: bool = False
        if num_slices > 1 and indices[1]<indices[0]:
            apply_backwards = True
        elif num_slices == 1:
            if indices[0] > 0 and len(annotations[indices[0]-1]) > 0:
                apply_backwards = False
            elif indices[0] < self.n_frames -1 and len(annotations[indices[0]+1]) > 0:
                apply_backwards = True

        # For every slice ...
        for i, index in enumerate(indices):
            if slices.ndim > 2:
                theimage = slices[index].copy()
            elif slices.ndim == 2 and index == 0:
                theimage = slices.copy()

            if annotations is not None:
                print(f"Slice index {index}, slice number {i}, rescue")
                if (index == 0 and not apply_backwards) or (index == self.n_frames-1 and apply_backwards):  # No previous mask to input
                    thepolygons = []
                    theids = []
                elif i == 0:  # Use polyline annotation made by user
                    prev_slice: int = index + 1 if apply_backwards else index - 1  # Use polyline annotation made by user
                    thepolygons = annotations[prev_slice]
                    theids = ids[prev_slice]
                else:  # Use the previous classifier output
                    thepolygons = []
                    theids = []

                    prev_slice: int = index + 1 if apply_backwards else index - 1

                    for apoly, thisid in zip(self.object_arrays[prev_slice], self.object_ids[prev_slice]):
                        thispolygon = QtGui.QPolygonF()
                        for thepoint in apoly:
                            thispolygon.append(QtCore.QPointF(thepoint[0], thepoint[1]))
                        thepolygons.append(thispolygon)
                        theids.append(thisid)

                self.object_ids[index] = []
                self.object_arrays[index] = []
                self.prob_arrays[index] = []
                for obj_num, [this_object, this_id] in enumerate(zip(thepolygons, theids)):
                    print(obj_num)
                    try:
                        this_mask = rimutils.mask_from_polylines(imsize=theimage.shape,
                                                                 polylines=[this_object],
                                                                 brushsz=0)
                    except IndexError:  # Polyline not big enough to make a mask from, skip this polyline
                        continue

                    this_object_array, this_prob_array = self.image_classifier.predict(theimage, this_mask)
                    if len(this_object_array) > 0:
                        self.object_arrays[index].append(this_object_array[0])
                        self.prob_arrays[index].append(this_prob_array)
                        self.object_ids[index].append(this_id)
                self.object_arrays[index] = numpy.asarray(self.object_arrays[index], dtype=object)

            else:
                return False

            if progress_signal is not None:
                progress_signal.emit(int((100 * (i + 1)) / num_slices))
        return True

    def reset_objects(self, n_frames: int) -> bool:
        self.n_frames = n_frames
        self.object_arrays = [numpy.empty((1, 0)) for i in range(self.n_frames)]
        self.prob_arrays = [numpy.empty((1, 0)) for i in range(self.n_frames)]
        self.object_ids = [numpy.empty((1, 0)) for i in range(self.n_frames)]

        return True



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from datetime import datetime
import os
from typing import Optional

import nbformat as nbf
import numpy
from PyQt6 import QtWidgets

import pyjamas.dialogs as dialogs
from pyjamas.pjscore import PyJAMAS
from pyjamas.pjsthreads import ThreadSignals
from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.rimage.rimml.rimclassifier import rimclassifier
import pyjamas.rimage.rimml.rimlr as rimlr
import pyjamas.rimage.rimml.rimsvm as rimsvm
import pyjamas.rimage.rimml.rimunet as rimunet
import pyjamas.rimage.rimml.rimrescunet as rimrescunet
import pyjamas.rimage.rimml.rimquickienet as rimquickienet
from pyjamas.rimage.rimml.rimneuralnet import rimneuralnet
from pyjamas.rutils import RUtils
from pyjamas.rimage.rimml.classifier_types import classifier_types


class RCBClassifiers(RCallback):
    COLAB_NOTEBOOK_APPENDIX: str = '_colab_notebook'

    def cbCreateLR(self, parameters: Optional[dict] = None, wait_for_thread: bool = False) -> bool:
        """
        Create a logistic regression classifier.

        :param parameters: dictionary containing the parameters to create a logistic regression classifier; a dialog opens if this parameter is set to None; keys are:

            ``positive_training_folder``:
                path to the folder containing positive training images, formatted as a string
            ``negative_training_folder``:
                path to the folder containing negative training images, formatted as a string
            ``hard_negative_training_folder``:
                path to the folder containing hard negative training images, formatted as a string
            ``histogram_of_gradients``:
                use the distribution of gradient orientations as image features, True or False
            ``train_image_size``:
                the number of rows and columns in the positive and negative training images, formatted as a tuple of two integers
            ``step_sz``:
                number of pixel rows and columns to skip when scanning test images for target structures, formatted as a tuple of two integers
            ``misclass_penalty_C``:
                penalty for misclassification of training samples, formatted as a float
        :param wait_for_thread: True if PyJAMAS must wait for the thread running this operation to complete, False otherwise.
        :return: True if the classifier was successfully created, False otherwise.
        """
        continue_flag = True

        if parameters is None or parameters is False:
            dialog = QtWidgets.QDialog()
            ui = dialogs.logregression.LRDialog()
            ui.setupUi(dialog)

            dialog.exec()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
            parameters = ui.parameters()

            dialog.close()

        if continue_flag:
            self.pjs.batch_classifier.image_classifier = rimlr.lr(parameters)
            self.launch_thread(self.pjs.batch_classifier.fit, {'stop': True}, finished_fn=self.finished_fn,
                               stop_fn=self.stop_fn, wait_for_thread=wait_for_thread)

            return True

        else:
            return False

    def cbCreateUNet(self, parameters: Optional[dict] = None, wait_for_thread: bool = False) -> bool:  # Handle IO errors.
        """
        Create a convolutional neural network with UNet architecture.

        :param parameters: dictionary containing the parameters to create a UNet; a dialog opens if this parameter is set to None; keys are:

            ``positive_training_folder``:
                path to the folder containing positive training images, formatted as a string
            ``train_image_size``:
                the number of rows and columns in the network input (train images will be scaled to this size) formatted as a tuple of two integers, both of the integers must be divisible by 16.
            ``step_sz``:
                number of pixel rows and columns to divide test images into, each subimage will be scaled to the network input size and processed, formatted as a tuple of two integers
            ``epochs``:
                maximum number of iterations over the training data, as an int
            ``learning_rate``:
                step size when updating the weights, as a float
            ``mini_batch_size``:
                size of mini batches, as an int
            ``erosion_width``:
                width of the erosion kernel to apply to the labeled image produced by the UNet, to separate touching objects, as an int
            ``generate_notebook``:
                whether a Jupyter notebook to create and train the UNet (e.g. in Google Colab) should be generated, as a bool (if True, the UNet will NOT be created)
            ``notebook_path``:
                where to store the Jupyter notebook if it must be created
        :param wait_for_thread: True if PyJAMAS must wait for the thread running this operation to complete, False otherwise.
        :return: True if the classifier was successfully created, False otherwise.
        """
        continue_flag = True

        if parameters is None or parameters is False:
            dialog = QtWidgets.QDialog()
            ui = dialogs.neuralnet.NeuralNetDialog()
            ui.setupUi(dialog)

            dialog.exec()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
            parameters = ui.parameters()

            dialog.close()

        if continue_flag:
            self.pjs.batch_classifier.image_classifier = rimunet.UNet(parameters)

            if not parameters.get('generate_notebook'):
                self.launch_thread(self.pjs.batch_classifier.fit, {'stop': True}, finished_fn=self.finished_fn,
                                   stop_fn=self.stop_fn, wait_for_thread=wait_for_thread)
            else:
                self._generate_neuralnet_notebook(parameters, classifier_types.UNET.value)

            return True

        else:
            return False

    def cbCreateReSCUNet(self, parameters: Optional[dict] = None, wait_for_thread: bool = False) -> bool:  # Handle IO errors.
        """
        Create a convolutional neural network with ReSCUNet architecture.

        :param parameters: dictionary containing the parameters to create a ReSCUNet; a dialog opens if this parameter is set to None; keys are:

            ``positive_training_folder``:
                path to the folder containing positive training images, formatted as a string
            ``train_image_size``:
                the number of rows and columns in the network input (train images will be scaled to this size) formatted as a tuple of two integers, both of the integers must be divisible by 16.
            ``step_sz``:
                number of pixel rows and columns to divide test images into, each subimage will be scaled to the network input size and processed, formatted as a tuple of two integers
            ``epochs``:
                maximum number of iterations over the training data, as an int
            ``learning_rate``:
                step size when updating the weights, as a float
            ``mini_batch_size``:
                size of mini batches, as an int
            ``erosion_width``:
                width of the erosion kernel to apply to the labeled image produced by the UNet, to separate touching objects, as an int
            ``concatenation_depth``:
                number of encoder blocks before previous segmentation mask and current image frame input streams are combined in the network
            ``generate_notebook``:
                whether a Jupyter notebook to create and train the UNet (e.g. in Google Colab) should be generated, as a bool (if True, the UNet will NOT be created)
            ``notebook_path``:
                where to store the Jupyter notebook if it must be created
            ``save_folder``:
                where to store resized images and weight maps, if empty the resized images and weight maps will not be saved
            ``resize_images_flag``:
                whether or not to resize images, resized images and weight_maps will be loaded from positive_training_folder if False
        :param wait_for_thread: True if PyJAMAS must wait for the thread running this operation to complete, False otherwise.
        :return: True if the classifier was successfully created, False otherwise.
        """
        continue_flag = True

        if parameters is None or parameters is False:
            dialog = QtWidgets.QDialog()
            ui = dialogs.rescuneuralnet.ReSCUNetDialog()
            ui.setupUi(dialog)

            dialog.exec()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
            parameters = ui.parameters()

            dialog.close()

        if continue_flag:
            self.pjs.batch_classifier.image_classifier = rimrescunet.ReSCUNet(parameters)

            if not parameters.get('generate_notebook'):
                self.launch_thread(self.pjs.batch_classifier.fit, {'stop': True}, finished_fn=self.finished_fn,
                                   stop_fn=self.stop_fn, wait_for_thread=wait_for_thread)
            else:
                self._generate_neuralnet_notebook(parameters, classifier_types.RESCUNET.value)

            return True

        else:
            return False

    def cbCreateQuickieNet(self, parameters: Optional[dict] = None, wait_for_thread: bool = False) -> bool:  # Handle IO errors.
        """
        Create a convolutional neural network with ReSCUNet architecture but training on entire images, not individual objects.
        Much faster (and probably more inaccurate) than ReSCUNet.

        :param parameters: dictionary containing the parameters to create a ReSCUNet; a dialog opens if this parameter is set to None; keys are:

            ``positive_training_folder``:
                path to the folder containing positive training images, formatted as a string
            ``train_image_size``:
                the number of rows and columns in the network input (train images will be scaled to this size) formatted as a tuple of two integers, both of the integers must be divisible by 16.
            ``step_sz``:
                number of pixel rows and columns to divide test images into, each subimage will be scaled to the network input size and processed, formatted as a tuple of two integers
            ``epochs``:
                maximum number of iterations over the training data, as an int
            ``learning_rate``:
                step size when updating the weights, as a float
            ``mini_batch_size``:
                size of mini batches, as an int
            ``erosion_width``:
                width of the erosion kernel to apply to the labeled image produced by the UNet, to separate touching objects, as an int
            ``concatenation_depth``:
                number of encoder blocks before previous segmentation mask and current image frame input streams are combined in the network
            ``generate_notebook``:
                whether a Jupyter notebook to create and train the UNet (e.g. in Google Colab) should be generated, as a bool (if True, the UNet will NOT be created)
            ``notebook_path``:
                where to store the Jupyter notebook if it must be created
            ``save_folder``:
                where to store resized images and weight maps, if empty the resized images and weight maps will not be saved
            ``resize_images_flag``:
                whether or not to resize images, resized images and weight_maps will be loaded from positive_training_folder if False
        :param wait_for_thread: True if PyJAMAS must wait for the thread running this operation to complete, False otherwise.
        :return: True if the classifier was successfully created, False otherwise.
        """
        continue_flag = True

        if parameters is None or parameters is False:
            dialog = QtWidgets.QDialog()
            ui = dialogs.rescuneuralnet.ReSCUNetDialog()
            ui.setupUi(dialog)

            dialog.exec()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
            parameters = ui.parameters()

            dialog.close()

        if continue_flag:
            self.pjs.batch_classifier.image_classifier = rimquickienet.QuickieNet(parameters)

            if not parameters.get('generate_notebook'):
                self.launch_thread(self.pjs.batch_classifier.fit, {'stop': True}, finished_fn=self.finished_fn,
                                   stop_fn=self.stop_fn, wait_for_thread=wait_for_thread)
            else:
                self._generate_neuralnet_notebook(parameters, classifier_types.QUICKIENET.value)

            return True

        else:
            return False

    def _generate_neuralnet_notebook(self, parameters: dict, architecture: classifier_types = classifier_types.UNET.value) -> bool:
        # Follow scheme of path generation from measure notebook from rcbbatchprocess._save_notebook
        path = parameters.get('notebook_path')

        # Create filename
        thenow = datetime.now()
        filename = thenow.strftime(
            f"{thenow.year:04}{thenow.month:02}{thenow.day:02}_{thenow.hour:02}{thenow.minute:02}{thenow.second:02}")
        filepath = os.path.join(path, filename)
        fname = RUtils.set_extension(filepath+RCBClassifiers.COLAB_NOTEBOOK_APPENDIX, PyJAMAS.notebook_extension)

        if architecture == classifier_types.UNET.value:
            template_fname = os.path.join(self.pjs.folder, "notebooks", "unet.ipynb")
        elif architecture == classifier_types.RESCUNET.value:
            template_fname = os.path.join(self.pjs.folder, "notebooks", "rescunet.ipynb")
        elif architecture == classifier_types.QUICKIENET.value:
            template_fname = os.path.join(self.pjs.folder, "notebooks", "quickienet.ipynb")

        with open(template_fname, 'r') as f:
            template = nbf.read(f, as_version=4)

        for acell in template.cells:
            if acell.cell_type == 'code':
                acell.source = acell.source.replace('{{positive_training_folder}}', f"'{parameters.get('positive_training_folder')}'")
                acell.source = acell.source.replace('{{save_folder}}', f"'{parameters.get('save_folder')}'")
                acell.source = acell.source.replace('{{train_image_size}}', str(parameters.get('train_image_size')))
                acell.source = acell.source.replace('{{step_sz}}', str(parameters.get('step_sz')))
                acell.source = acell.source.replace('{{learning_rate}}', str(parameters.get('learning_rate')))
                acell.source = acell.source.replace('{{mini_batch_size}}', str(parameters.get('mini_batch_size')))
                acell.source = acell.source.replace('{{epochs}}', str(parameters.get('epochs')))
                acell.source = acell.source.replace('{{validation_split}}', str(parameters.get('validation_split')))
                acell.source = acell.source.replace('{{erosion_width}}', str(parameters.get('erosion_width')))
                acell.source = acell.source.replace('{{early_stopper}}', str(parameters.get('early_stopper')))
                acell.source = acell.source.replace('{{lr_scheduler}}', str(parameters.get('lr_scheduler')))
                acell.source = acell.source.replace('{{model_checkpoint}}', str(parameters.get('model_checkpoint')))
                acell.source = acell.source.replace('{{logging}}', str(parameters.get('logging')))
                acell.source = acell.source.replace('{{concatenation_level}}', str(parameters.get('concatenation_level')))

        with open(fname, 'w') as f:
            nbf.write(template, f)

        return True

    def cbCreateSVM(self, parameters: Optional[dict] = None, wait_for_thread: bool = False) -> bool:  # Handle IO errors.
        """
        Create a support vector machine classifier.

        :param parameters: dictionary containing the parameters to create a logistic regression classifier; a dialog opens if this parameter is set to None; keys are:

            ``positive_training_folder``:
                path to the folder containing positive training images, formatted as a string
            ``negative_training_folder``:
                path to the folder containing negative training images, formatted as a string
            ``hard_negative_training_folder``:
                path to the folder containing hard negative training images, formatted as a string
            ``histogram_of_gradients``:
                use the distribution of gradient orientations as image features, True or False
            ``train_image_size``:
                the number of rows and columns in the positive and negative training images, formatted as a tuple of two integers
            ``step_sz``:
                number of pixel rows and columns to skip when scanning test images for target structures, formatted as a tuple of two integers
            ``misclass_penalty_C``:
                penalty for misclassification of training samples, formatted as a float
            ``kernel_type``:
                type of kernel ('linear' or 'rbf')
        :param wait_for_thread: True if PyJAMAS must wait for the thread running this operation to complete, False otherwise.
        :return: True if the classifier was successfully created, False otherwise.
        """


        continue_flag = True

        if parameters is None or parameters is False:
            dialog = QtWidgets.QDialog()
            ui = dialogs.svm.SVMDialog()
            ui.setupUi(dialog)

            dialog.exec()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
            parameters = ui.parameters()

            dialog.close()

        if continue_flag:
            self.pjs.batch_classifier.image_classifier = rimsvm.svm(parameters)
            self.launch_thread(self.pjs.batch_classifier.fit, {'stop': True}, finished_fn=self.finished_fn,
                               stop_fn=self.stop_fn, wait_for_thread=wait_for_thread)

            return True

        else:
            return False

    def cbApplyClassifier(self, firstSlice: Optional[int] = None, lastSlice: Optional[int] = None,
                          wait_for_thread: bool = False) -> bool:    # Handle IO errors.
        """
        Apply the current classifier to detect objects in the open image.

        :param firstSlice: slice number for the first slice to use (minimum is 1); a dialog will open if this parameter is None.
        :param lastSlice: slice number for the last slice to use; a dialog will open if this parameter is None.
        :param wait_for_thread: True if PyJAMAS must wait for the thread running this operation to complete, False otherwise.
        :return: True if the classifier is applied, False if the process is cancelled.
        """
        if (firstSlice is False or firstSlice is None or lastSlice is False or lastSlice is None) and self.pjs is not None:
            dialog = QtWidgets.QDialog()
            ui = dialogs.timepoints.TimePointsDialog()

            lastSlice = 1 if self.pjs.n_frames == 1 else self.pjs.slices.shape[0]
            ui.setupUi(dialog, firstslice=self.pjs.curslice + 1, lastslice=lastSlice)

            dialog.exec()
            dialog.show()
            # If the dialog was closed by pressing OK, then run the measurements.
            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
            firstSlice, lastSlice = ui.parameters()

            dialog.close()
        else:
            continue_flag = True

        if continue_flag:

            if firstSlice <= lastSlice:
                theslicenumbers = numpy.arange(firstSlice - 1, lastSlice, dtype=int)
            else:
                increasing_order: bool = False if type(self.pjs.batch_classifier.image_classifier) == rimquickienet.QuickieNet else True

                if increasing_order:
                    theslicenumbers = numpy.arange(lastSlice - 1, firstSlice, dtype=int)
                else:
                    theslicenumbers = numpy.arange(firstSlice - 1, lastSlice - 2, -1, dtype=int)

            self.launch_thread(self.apply_classifier, {'theslices': theslicenumbers, 'progress': True, 'stop': True},
                               finished_fn=self.finished_fn,  progress_fn=self.progress_fn, stop_fn=self.stop_fn,
                               wait_for_thread=wait_for_thread)

            return True
        else:
            return False

    def apply_classifier(self, theslices: numpy.ndarray, progress_signal: ThreadSignals,
                         stop_signal: ThreadSignals) -> bool:
        # Make sure that the slices are in a 1D numpy array.
        theslices = numpy.atleast_1d(theslices)

        if stop_signal is not None:
            stop_signal.emit("Applying classifier ...")

        if type(self.pjs.batch_classifier.image_classifier) in [rimrescunet.ReSCUNet, rimquickienet.QuickieNet]:
            self.pjs.batch_classifier.predict(self.pjs.slices, theslices, progress_signal, self.pjs.polylines,
                                              self.pjs.polyline_ids)
        else:
            self.pjs.batch_classifier.predict(self.pjs.slices, theslices, progress_signal)

        # For every slice ...
        for index in theslices:
            if type(self.pjs.batch_classifier.image_classifier) in [rimlr.lr, rimsvm.svm]:
                self.add_classifier_boxes(self.pjs.batch_classifier.box_arrays[index], index, False)
            elif type(self.pjs.batch_classifier.image_classifier) in [rimrescunet.ReSCUNet, rimquickienet.QuickieNet]:
                self.add_neuralnet_polylines(self.pjs.batch_classifier.object_arrays[index],
                                             self.pjs.batch_classifier.object_ids[index], index, False)
            elif type(self.pjs.batch_classifier.image_classifier) is rimunet.UNet:
                self.add_neuralnet_polylines(self.pjs.batch_classifier.object_arrays[index], slice_index=index,
                                             paint=False)
            elif type(self.pjs.batch_classifier.image_classifier) in [rimrescunet.ReSCUNet, rimquickienet.QuickieNet]:
                self.add_neuralnet_polylines(self.pjs.batch_classifier.object_arrays[index],
                                             self.pjs.batch_classifier.object_ids[index], index, False)
            else:
                self.pjs.statusbar.showMessage(f"Wrong classifier type.")
                return False

        return True

    def add_neuralnet_polylines(self, polylines: Optional[numpy.ndarray] = None, ids: Optional[numpy.ndarray] = None,
                                slice_index: Optional[int] = None, paint: bool = True) -> bool:
        if polylines is None or polylines is False or not hasattr(polylines, "size") or polylines.size == 0:
            return False

        if slice_index is None or slice_index is False:
            slice_index = self.pjs.curslice

        for p, aPoly in enumerate(polylines):
            if ids is None:
                self.pjs.addPolyline(aPoly, slice_index, paint=paint)
            else:
                self.pjs.addPolyline(aPoly, slice_index, theid=ids[p], paint=paint)

        return True

    def add_classifier_boxes(self, boxes: Optional[numpy.ndarray] = None, slice_index: Optional[int] = None,
                             paint: bool = True) -> bool:  # The first slice_index should be 0.
        if boxes is None or boxes is False or boxes == []:
            return False

        if slice_index is None or slice_index is False:
            slice_index = self.pjs.curslice

        for aBox in boxes:
            # Boxes stored as [minrow, mincol, maxrow, maxcol]
            self.pjs.addPolyline([[aBox[1], aBox[0]], [aBox[3], aBox[0]], [aBox[3], aBox[2]],
                                  [aBox[1], aBox[2]], [aBox[1], aBox[0]]], slice_index, paint=paint)

        return True

    def cbNonMaxSuppression(self, parameters: Optional[dict] = None, firstSlice: Optional[int] = None,
                            lastSlice: Optional[int] = None) -> bool:
        """
        Apply non-maximum suppression to remove redundant objects from an image.

        :param parameters: dictionary containing the parameters for non-maximum suppression; a dialog will open if this parameter is None; keys are:

            ``prob_threshold``:
                lower threshold for the probability that a detected object represents an instance of the positive training set (returned by the classifier), as a float
            ``iou_threshold``:
                maximum value for the intersection-over-union ratio for the area of two detected objects, as a float; 0.0 prevents any overlaps between objects, 1.0 allows full overlap
            ``max_num_objects``:
                maximum number of objects present in the image, as an integer; objects will be discarded from lowest to highest probability of the object representing an instance of the positive training set
        :param firstSlice: slice number for the first slice to use (minimum is 1); a dialog will open if this parameter is None.
        :param lastSlice: slice number for the last slice to use; a dialog will open if this parameter is None.
        :return: True if non-maximum suppression is applied, False if the process is cancelled.
        """

        if self.pjs.batch_classifier is None or type(self.pjs.batch_classifier.image_classifier) is rimunet.UNet:

            return False

        continue_flag = True

        if parameters is None or parameters is False:
            dialog = QtWidgets.QDialog()
            ui = dialogs.nonmax_suppr.NonMaxDialog(self.pjs)
            ui.setupUi(dialog)
            dialog.exec()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted

            if continue_flag:
                parameters = ui.parameters()

            dialog.close()

        if not continue_flag:
            return False

        if (firstSlice is None or firstSlice is False) and (lastSlice is None or lastSlice is False):
            dialog = QtWidgets.QDialog()
            ui = dialogs.timepoints.TimePointsDialog()
            ui.setupUi(dialog, dialogs.timepoints.TimePointsDialog.firstSlice,
                       dialogs.timepoints.TimePointsDialog.lastSlice)

            dialog.exec()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted

            if continue_flag:
                firstSlice, lastSlice = ui.parameters()

            dialog.close()

        if firstSlice <= lastSlice:
            theslicenumbers = numpy.arange(firstSlice - 1, lastSlice, dtype=int)
        else:
            theslicenumbers = numpy.arange(lastSlice - 1, firstSlice, dtype=int)

        self.pjs.batch_classifier.non_max_suppression(
            parameters.get('prob_threshold', rimclassifier.DEFAULT_PROB_THRESHOLD),
            parameters.get('iou_threshold', rimclassifier.DEFAULT_IOU_THRESHOLD),
            parameters.get('max_num_objects', rimclassifier.DEFAULT_MAX_NUM_OBJECTS),
            theslicenumbers
        )

        for index in theslicenumbers:
            self.pjs.annotations.cbDeleteSliceAnn(index)
            self.pjs.classifiers.add_classifier_boxes(self.pjs.batch_classifier.box_arrays[index][self.pjs.batch_classifier.good_box_indices[index]], index, True)

        self.pjs.repaint()

        return True

    def cbSetNeuralNetStepSize(self, step_sz: tuple = None) -> bool:
        """
        Set the step size for tiling over an input image when applying a convolutional neural network.
        If step_sz >= train_image_size no tiling occurs.

        :param step_sz: the step size (int, int) for tiling in the vertical and horizontal directions.
        :return: True if option is succesfully set, False otherwise.
        """
        ok_flag = True
        #print(type(self.pjs.batch_classifier.image_classifier))
        if not issubclass(type(self.pjs.batch_classifier.image_classifier), rimneuralnet):
            print("Classifier must be a neural network.")
            return False

        if step_sz is None or not step_sz:
            # Read user input for step size and then prediction image size.
            step_sz_str, ok_flag = QtWidgets.QInputDialog.getText(None, 'Step Size',
                                                                   '(row, column) step size: ',
                                                                   QtWidgets.QLineEdit.EchoMode.Normal,
                                                                   f"{self.pjs.batch_classifier.image_classifier.step_sz}")
        if ok_flag:
            try:
                step_sz = eval(step_sz_str)
            except:
                print(f"Incorrect input tuple.")
                ok_flag = False

        if ok_flag:
            self.pjs.batch_classifier.image_classifier.step_sz = step_sz
            return True
        else:
            return False



from typing import List

import matplotlib.pyplot as plt
import numpy
import pandas as pd

from pyjamas.rannotations.rpolyline import RPolyline
from pyjamas.rplugins.base import PJSPluginABC

class PJSPlugin(PJSPluginABC):
    def name(self) -> str:
        return 'Heterogeneity'

    def run(self, parameters: dict) -> bool:
        # Array with an index per slice in the image stack.
        slices = numpy.arange(self.pjs.n_frames, dtype=numpy.int16)

        # Find the maximum number of polylines in a slice.
        max_n_polylines = 0
        for i in slices:
            max_n_polylines = max(max_n_polylines, len(self.pjs.polylines[i]))

        # Create a pandas data frame to store the measurements.
        # The data frame will have one column per polyline, and one row per slice in the image.
        n_columns: int = max_n_polylines
        n_rows: int = slices.shape[0]

        column_names: List[str] = ['heterogeneity_' + str(i) for i in range(1, max_n_polylines+1)]

        heterogeneity_df: pd.DataFrame = pd.DataFrame(numpy.nan * numpy.zeros((n_rows, n_columns)), columns=column_names, index=slices+1)

        # For every slice ...
        for i in slices:
            theimage = self.pjs.slices[i]

            # Retrieve the polylines in this slice.
            polygon_slice = self.pjs.polylines[i]

            n_polylines = len(polygon_slice)

            # For every polyline ...
            for j in range(n_polylines):
                # We take advantage of the RPolyline class in PyJAMAS, which contains methods to measure
                # morphology and pixel values for the polyline.
                thepolyline = RPolyline(polygon_slice[j])

                intensities = thepolyline.pixel_values(theimage, self.pjs.brush_size)
                mean_poly = intensities[0]
                std_poly = intensities[2]
                heterogeneity_df.loc[i + 1, 'heterogeneity_' + str(j + 1)] = std_poly / mean_poly

        # Plot results.
        heterogeneity_df.plot()
        plt.xlabel('slice index')
        plt.ylabel('heterogeneity')
        plt.show()

        # Print full results in the terminal.
        with pd.option_context('display.max_rows', None, 'display.max_columns', None):
            print(heterogeneity_df)

        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PyQt6 import QtCore, QtWidgets

from pyjamas.pjscore import PyJAMAS
from pyjamas.rimage.rimutils import rimutils


class FindPunctaDialog(object):

    first_slice: int = -1
    last_slice: int = -1
    mean_filter_width: int = -1
    difference_threshold: float = -1.
    max_size_merge: int = -1

    def __init__(self, pjs: PyJAMAS):
        super().__init__()

        self.pjs = pjs

    def setupUi(self, Dialog, first_slice: int = None, last_slice: int = None, mean_filter_width: int = None,
                difference_threshold: float = None, max_size_merge: int = None):
        if FindPunctaDialog.first_slice < 0:
            FindPunctaDialog.first_slice = first_slice or 0

        if FindPunctaDialog.last_slice < 0:
            FindPunctaDialog.last_slice = last_slice or 1

        if FindPunctaDialog.mean_filter_width < 1:
            FindPunctaDialog.mean_filter_width = rimutils.WATER_HPFILTER_WIDTH
        else:
            FindPunctaDialog.mean_filter_width = mean_filter_width

        if FindPunctaDialog.difference_threshold < 0:
            FindPunctaDialog.difference_threshold = rimutils.WATER_POSTHPFILTER_THRESHOLD
        else:
            FindPunctaDialog.difference_threshold = difference_threshold

        if FindPunctaDialog.max_size_merge < 1:
            FindPunctaDialog.max_size_merge = rimutils.WATER_MAX_SIZE_MERGE
        else:
            FindPunctaDialog.max_size_merge = max_size_merge

        self.dialog = Dialog
        Dialog.setObjectName("Dialog")
        Dialog.resize(609, 206)
        self.ok_buttonbox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.ok_buttonbox.setGeometry(QtCore.QRect(230, 160, 341, 32))
        self.ok_buttonbox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.ok_buttonbox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.ok_buttonbox.setObjectName("ok_buttonbox")
        self.input_label = QtWidgets.QLabel(parent=Dialog)
        self.input_label.setGeometry(QtCore.QRect(30, 70, 111, 16))
        self.input_label.setObjectName("input_label")
        self.first_slice_spinbox = QtWidgets.QSpinBox(parent=Dialog)
        self.first_slice_spinbox.setGeometry(QtCore.QRect(105, 28, 104, 26))
        self.first_slice_spinbox.setMinimum(1)
        self.first_slice_spinbox.setMaximum(last_slice)
        self.first_slice_spinbox.setValue(first_slice)
        self.first_slice_spinbox.setObjectName("first_slice_spinbox")
        self.last_slice_spinbox = QtWidgets.QSpinBox(parent=Dialog)
        self.last_slice_spinbox.setGeometry(QtCore.QRect(286, 26, 104, 26))
        self.last_slice_spinbox.setMinimum(1)
        self.last_slice_spinbox.setMaximum(last_slice)
        self.last_slice_spinbox.setValue(first_slice)
        self.last_slice_spinbox.setObjectName("last_slice_spinbox")
        self.first_slice_label = QtWidgets.QLabel(parent=Dialog)
        self.first_slice_label.setGeometry(QtCore.QRect(35, 30, 101, 16))
        self.first_slice_label.setObjectName("first_slice_label")
        self.last_slice_label = QtWidgets.QLabel(parent=Dialog)
        self.last_slice_label.setGeometry(QtCore.QRect(220, 30, 101, 16))
        self.last_slice_label.setObjectName("last_slice_label")
        self.mean_width_slider = QtWidgets.QSlider(parent=Dialog)
        self.mean_width_slider.setGeometry(QtCore.QRect(170, 67, 341, 22))
        self.mean_width_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.mean_width_slider.setMinimum(2)
        self.mean_width_slider.setMaximum(rimutils.WATER_HPFILTER_WIDTH_MAX)
        self.mean_width_slider.setValue(FindPunctaDialog.mean_filter_width)
        self.mean_width_slider.valueChanged.connect(lambda: (self.mean_width_edit.setText(str(self.mean_width_slider.value()))))
        self.mean_width_slider.sliderReleased.connect(self._update_mean_width_slider)
        self.mean_width_slider.setObjectName("mean_width_slider")
        self.diff_threshold_slider = QtWidgets.QSlider(parent=Dialog)
        self.diff_threshold_slider.setGeometry(QtCore.QRect(170, 97, 341, 22))
        self.diff_threshold_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.diff_threshold_slider.setMinimum(1)
        self.diff_threshold_slider.setMaximum(self.pjs.imagedata.max()-self.pjs.imagedata.min())
        self.diff_threshold_slider.setValue(FindPunctaDialog.difference_threshold)
        self.diff_threshold_slider.valueChanged.connect(
            lambda: (self.diff_threshold_edit.setText(str(self.diff_threshold_slider.value()))))
        self.diff_threshold_slider.sliderReleased.connect(self._update_diff_threshold_slider)
        self.diff_threshold_slider.setObjectName("diff_threshold_slider")
        self.input_label_2 = QtWidgets.QLabel(parent=Dialog)
        self.input_label_2.setGeometry(QtCore.QRect(30, 100, 131, 16))
        self.input_label_2.setObjectName("input_label_2")
        self.input_label_3 = QtWidgets.QLabel(parent=Dialog)
        self.input_label_3.setGeometry(QtCore.QRect(30, 130, 131, 16))
        self.input_label_3.setObjectName("input_label_3")
        self.max_size_slider = QtWidgets.QSlider(parent=Dialog)
        self.max_size_slider.setGeometry(QtCore.QRect(170, 127, 341, 22))
        self.max_size_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.max_size_slider.setMinimum(1)
        self.max_size_slider.setMaximum(rimutils.WATER_MAX_SIZE_MERGE_MAX)
        self.max_size_slider.setValue(FindPunctaDialog.max_size_merge)
        self.max_size_slider.valueChanged.connect(
            lambda: (self.max_size_edit.setText(str(self.max_size_slider.value()))))
        self.max_size_slider.sliderReleased.connect(self._update_max_size_slider)
        self.max_size_slider.setObjectName("max_size_slider")
        self.mean_width_edit = QtWidgets.QLineEdit(parent=Dialog)
        self.mean_width_edit.setGeometry(QtCore.QRect(521, 68, 61, 21))
        self.mean_width_edit.textEdited.connect(self._set_new_mean_width_slider_value)
        self.mean_width_edit.editingFinished.connect(self._update_mean_width_edit)
        self.mean_width_edit.setText(str(FindPunctaDialog.mean_filter_width))
        self.mean_width_edit.setObjectName("mean_width_edit")
        self.diff_threshold_edit = QtWidgets.QLineEdit(parent=Dialog)
        self.diff_threshold_edit.setGeometry(QtCore.QRect(520, 100, 61, 21))
        self.diff_threshold_edit.setText(str(FindPunctaDialog.difference_threshold))
        self.diff_threshold_edit.textEdited.connect(self._set_new_diff_threshold_slider_value)
        self.diff_threshold_edit.editingFinished.connect(self._update_diff_threshold_edit)
        self.diff_threshold_edit.setObjectName("diff_threshold_edit")
        self.max_size_edit = QtWidgets.QLineEdit(parent=Dialog)
        self.max_size_edit.setGeometry(QtCore.QRect(520, 130, 61, 21))
        self.max_size_edit.setText(str(FindPunctaDialog.max_size_merge))
        self.max_size_edit.textEdited.connect(self._set_new_max_size_slider_value)
        self.max_size_edit.editingFinished.connect(self._update_max_size_edit)
        self.max_size_edit.setObjectName("max_size_edit")

        self.retranslateUi(Dialog)
        self.ok_buttonbox.accepted.connect(self.accept)
        self.ok_buttonbox.rejected.connect(self.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

        self._old_annotations = self.pjs.copy_annotations()
        self._clean_n_run_water()

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Find puncta"))
        self.input_label.setText(_translate("Dialog", "mean filter width"))
        self.first_slice_label.setText(_translate("Dialog", "first slice"))
        self.last_slice_label.setText(_translate("Dialog", "last slice"))
        self.input_label_2.setText(_translate("Dialog", "difference threshold"))
        self.input_label_3.setText(_translate("Dialog", "maximum merge size"))

    def parameters(self):
        FindPunctaDialog.first_slice = self.first_slice_spinbox.value()
        FindPunctaDialog.last_slice = self.last_slice_spinbox.value()
        FindPunctaDialog.mean_filter_width = self.mean_width_slider.value()
        FindPunctaDialog.difference_threshold = self.diff_threshold_slider.value()
        FindPunctaDialog.max_size_merge = self.max_size_slider.value()

        return {
            'first_slice': FindPunctaDialog.first_slice,
            'last_slice': FindPunctaDialog.last_slice,
            'mean_filter_width': FindPunctaDialog.mean_filter_width,
            'difference_threshold': FindPunctaDialog.difference_threshold,
            'max_size_merge': FindPunctaDialog.max_size_merge,
        }

    def _update_max_size_slider(self):
        self.max_size_edit.setText(str(self.max_size_slider.value()))

        self._clean_n_run_water()

        return True
    
    def _set_new_max_size_slider_value(self):
        try:
            self.max_size_slider.setValue(int(self.mean_width_edit.text()))
        except ValueError:
            pass

        return True

    def _update_max_size_edit(self):
        self._set_new_max_size_slider_value()
        self._clean_n_run_water()

        return True

    def _update_mean_width_slider(self):
        self.mean_width_edit.setText(str(self.mean_width_slider.value()))

        self._clean_n_run_water()

        return True

    def _set_new_mean_width_slider_value(self):
        try:
            self.mean_width_slider.setValue(int(self.mean_width_edit.text()))
        except ValueError:
            pass

        return True

    def _update_mean_width_edit(self):
        self._set_new_mean_width_slider_value()
        self._clean_n_run_water()

        return True

    def _update_diff_threshold_slider(self):
        self.diff_threshold_edit.setText(str(self.diff_threshold_slider.value()))

        self._clean_n_run_water()

        return True

    def _set_new_diff_threshold_slider_value(self):
        try:
            self.diff_threshold_slider.setValue(int(self.diff_threshold_edit.text()))
        except ValueError:
            pass

        return True

    def _update_diff_threshold_edit(self):
        self._set_new_diff_threshold_slider_value()
        self._clean_n_run_water()

        return True

    def _clean_n_run_water(self):
        self.pjs.annotations.cbDeleteAllAnn()
        self._water()

        return True

    def _water(self):
        parameters = self.parameters()

        self.pjs.image.cbFindPuncta(self.pjs.curslice+1, self.pjs.curslice+1, parameters.get('mean_filter_width'),
                                    parameters.get('difference_threshold'),
                                    parameters.get('max_size_merge'), True)

        return True

    def accept(self):
        self.pjs.annotations.cbDeleteAllAnn()
        self.pjs.paste_annotations(self._old_annotations)
        #self._water()
        self.dialog.accept()

    def reject(self):
        self.pjs.annotations.cbDeleteAllAnn()
        self.pjs.paste_annotations(self._old_annotations)
        self.dialog.reject()



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import inspect
from functools import partial
from typing import Optional, Tuple

from PyQt6 import QtCore, QtWidgets

from pyjamas.rimage.rimml.rimrescunet import ReSCUNet
from pyjamas.rutils import RUtils


class ReSCUNetDialog:
    positive_training_folder: str = ''
    save_folder: str = ''
    train_image_size: Tuple[int, int] = (0, 0)
    epochs: int = ReSCUNet.EPOCHS
    learning_rate: float = ReSCUNet.LEARNING_RATE
    mini_batch_size: int = ReSCUNet.BATCH_SIZE
    erosion_width: int = ReSCUNet.EROSION_WIDTH
    step_sz: Tuple[int, int] = ReSCUNet.STEP_SIZE
    validation_split: float = ReSCUNet.VALIDATION_SPLIT
    concatenation_level: int = ReSCUNet.CONCATENATION_LEVEL
    early_stopper: dict = ReSCUNet.EARLY_STOPPER
    lr_scheduler: dict = ReSCUNet.LR_SCHEDULER
    model_checkpoint: dict = ReSCUNet.MODEL_CHECKPOINT
    logging: dict = ReSCUNet.LOGGING

    generate_notebook: bool = False
    notebook_path: str = ''

    def __init__(self):
        super().__init__()

    def setupUi(self, NNet: QtWidgets.QDialog, parameters: Optional[dict] = None):
        attributes = inspect.getmembers(ReSCUNetDialog, lambda a: not (inspect.isroutine(a)))
        attributes = [a for a in attributes if not (a[0].startswith('__') and a[0].endswith('__'))]
        for key, value in attributes:
            if parameters is not None and parameters is not False and key in parameters:
                ReSCUNetDialog.__setattr__(key, parameters[key])

        NNet.setObjectName("NNet")
        NNet.resize(614, 565)

        # ===================================== GroupBox for Folders ===================================================

        self.groupBox_2 = QtWidgets.QGroupBox(NNet)
        self.groupBox_2.setObjectName("groupBox_2")
        self.groupBox_2.setGeometry(30, 26, 551, 130)

        self.label_ptf = QtWidgets.QLabel(self.groupBox_2)
        self.label_ptf.setObjectName("label_ptf")
        self.label_ptf.setGeometry(31, 26, 141, 24)
        self.positive_training_folder_edit = QtWidgets.QLineEdit(self.groupBox_2)
        self.positive_training_folder_edit.setObjectName("positive_training_folder_edit")
        self.positive_training_folder_edit.setGeometry(220, 30, 261, 21)
        self.positive_training_folder_edit.setText(ReSCUNetDialog.positive_training_folder)
        self.btnSavePositive = QtWidgets.QToolButton(self.groupBox_2)
        self.btnSavePositive.setObjectName("btnSavePositive")
        self.btnSavePositive.setGeometry(490, 30, 26, 22)
        self.btnSavePositive.clicked.connect(self._open_positive_folder_dialog)

        self.label_sf = QtWidgets.QLabel(self.groupBox_2)
        self.label_sf.setObjectName("label_sf")
        self.label_sf.setGeometry(31, 56, 141, 24)
        self.save_folder_edit = QtWidgets.QLineEdit(self.groupBox_2)
        self.save_folder_edit.setObjectName("save_folder_edit")
        self.save_folder_edit.setGeometry(220, 60, 261, 21)
        self.save_folder_edit.setText(ReSCUNetDialog.save_folder)
        self.btnSaveSave = QtWidgets.QToolButton(self.groupBox_2)
        self.btnSaveSave.setObjectName("btnSavePositive")
        self.btnSaveSave.setGeometry(490, 60, 26, 22)
        self.btnSaveSave.clicked.connect(self._open_save_folder_dialog)

        self.cbGenerateNotebook = QtWidgets.QCheckBox(self.groupBox_2)
        self.cbGenerateNotebook.setObjectName("cbGenerateNotebook")
        self.cbGenerateNotebook.setGeometry(31, 86, 200, 24)
        self.cbGenerateNotebook.setChecked(ReSCUNetDialog.generate_notebook)
        self.notebook_path_edit = QtWidgets.QLineEdit(self.groupBox_2)
        self.notebook_path_edit.setObjectName("notebook_path_edit")
        self.notebook_path_edit.setGeometry(220, 90, 261, 21)
        self.notebook_path_edit.setText(ReSCUNetDialog.notebook_path)
        self.notebook_path_edit.setEnabled(ReSCUNetDialog.generate_notebook)
        self.btnSaveNotebook = QtWidgets.QToolButton(self.groupBox_2)
        self.btnSaveNotebook.setObjectName("btnSaveNotebook")
        self.btnSaveNotebook.setGeometry(490, 90, 26, 22)
        self.btnSaveNotebook.clicked.connect(self._open_notebook_path_dialog)
        self.btnSaveNotebook.setEnabled(ReSCUNetDialog.generate_notebook)
        self.cbGenerateNotebook.stateChanged.connect(partial(self._connect_check, self.cbGenerateNotebook,
                                                             [self.notebook_path_edit, self.btnSaveNotebook],
                                                             True))

        self.label_ptf.setEnabled(not self.cbGenerateNotebook.isChecked())
        self.positive_training_folder_edit.setEnabled(not self.cbGenerateNotebook.isChecked())
        self.btnSavePositive.setEnabled(not self.cbGenerateNotebook.isChecked())
        self.label_sf.setEnabled(not self.cbGenerateNotebook.isChecked())
        self.save_folder_edit.setEnabled(not self.cbGenerateNotebook.isChecked())
        self.btnSaveSave.setEnabled(not self.cbGenerateNotebook.isChecked())
        self.cbGenerateNotebook.stateChanged.connect(partial(self._connect_check, self.cbGenerateNotebook,
                                                             [self.label_ptf, self.positive_training_folder_edit,
                                                              self.btnSavePositive, self.label_sf,
                                                              self.save_folder_edit, self.btnSaveSave],
                                                             False))

        # ======================== GroupBox for Network Input Size and Step Size  ======================================

        self.groupBox_3 = QtWidgets.QGroupBox(NNet)
        self.groupBox_3.setObjectName("groupBox_3")
        self.groupBox_3.setGeometry(30, 167, 251, 61)

        self.label_w = QtWidgets.QLabel(self.groupBox_3)
        self.label_w.setObjectName("label_w")
        self.label_w.setGeometry(31, 28, 141, 24)
        self.lnWidth = QtWidgets.QLineEdit(self.groupBox_3)
        self.lnWidth.setObjectName("lnWidth")
        self.lnWidth.setGeometry(70, 30, 31, 21)
        self.lnWidth.setText(str(ReSCUNetDialog.train_image_size[1]))
        self.label_h = QtWidgets.QLabel(self.groupBox_3)
        self.label_h.setObjectName("label_h")
        self.label_h.setGeometry(120, 28, 141, 24)
        self.lnHeight = QtWidgets.QLineEdit(self.groupBox_3)
        self.lnHeight.setObjectName("lnHeight")
        self.lnHeight.setGeometry(170, 30, 31, 21)
        self.lnHeight.setText(str(ReSCUNetDialog.train_image_size[0]))

        self.groupBox_6 = QtWidgets.QGroupBox(NNet)
        self.groupBox_6.setObjectName("groupBox_6")
        self.groupBox_6.setGeometry(330, 167, 251, 61)

        self.label_r = QtWidgets.QLabel(self.groupBox_6)
        self.label_r.setObjectName("label_r")
        self.label_r.setGeometry(31, 28, 141, 24)
        self.lnRow = QtWidgets.QLineEdit(self.groupBox_6)
        self.lnRow.setObjectName("lnRow")
        self.lnRow.setGeometry(70, 30, 31, 21)
        self.lnRow.setText(str(ReSCUNetDialog.step_sz[0]))
        self.label_c = QtWidgets.QLabel(self.groupBox_6)
        self.label_c.setObjectName("label_c")
        self.label_c.setGeometry(120, 28, 141, 24)
        self.lnColumn = QtWidgets.QLineEdit(self.groupBox_6)
        self.lnColumn.setObjectName("lnColumn")
        self.lnColumn.setGeometry(180, 30, 31, 21)
        self.lnColumn.setText(str(ReSCUNetDialog.step_sz[1]))

        # ================================ GroupBox for Other Parameters ===============================================

        self.groupBox_5 = QtWidgets.QGroupBox(NNet)
        self.groupBox_5.setObjectName("groupBox_5")
        self.groupBox_5.setGeometry(30, 249, 551, 100)

        self.label_lr = QtWidgets.QLabel(self.groupBox_5)
        self.label_lr.setObjectName("label_lr")
        self.label_lr.setGeometry(31, 31, 141, 16)
        self.lnEta = QtWidgets.QLineEdit(self.groupBox_5)
        self.lnEta.setObjectName("lnEta")
        self.lnEta.setGeometry(122, 27, 46, 21)
        self.lnEta.setText(str(ReSCUNetDialog.learning_rate))
        self.label_bs = QtWidgets.QLabel(self.groupBox_5)
        self.label_bs.setObjectName("label_bs")
        self.label_bs.setGeometry(211, 31, 95, 16)
        self.lnBatchSz = QtWidgets.QLineEdit(self.groupBox_5)
        self.lnBatchSz.setObjectName("lnBatchSz")
        self.lnBatchSz.setGeometry(276, 27, 46, 21)
        self.lnBatchSz.setText(str(ReSCUNetDialog.mini_batch_size))
        self.label_epochs = QtWidgets.QLabel(self.groupBox_5)
        self.label_epochs.setObjectName("label_epochs")
        self.label_epochs.setGeometry(376, 31, 45, 16)
        self.lnEpochs = QtWidgets.QLineEdit(self.groupBox_5)
        self.lnEpochs.setObjectName("lnEpochs")
        self.lnEpochs.setGeometry(426, 27, 41, 21)
        self.lnEpochs.setText(str(ReSCUNetDialog.epochs))

        self.label_val = QtWidgets.QLabel(self.groupBox_5)
        self.label_val.setObjectName("label_val")
        self.label_val.setGeometry(31, 75, 141, 16)
        self.lnValSplit = QtWidgets.QLineEdit(self.groupBox_5)
        self.lnValSplit.setObjectName("lnValSplit")
        self.lnValSplit.setGeometry(122, 71, 46, 21)
        self.lnValSplit.setText(str(ReSCUNetDialog.validation_split))
        self.label_ew = QtWidgets.QLabel(self.groupBox_5)
        self.label_ew.setObjectName("label_ew")
        self.label_ew.setGeometry(201, 75, 141, 16)
        self.lnErosionWidth = QtWidgets.QLineEdit(self.groupBox_5)
        self.lnErosionWidth.setObjectName("lnErosionWidth")
        self.lnErosionWidth.setGeometry(286, 71, 46, 21)
        self.lnErosionWidth.setText(str(ReSCUNetDialog.erosion_width))
        self.label_cat = QtWidgets.QLabel(self.groupBox_5)
        self.label_cat.setObjectName("label_cat")
        self.label_cat.setGeometry(346, 75, 120, 16)
        self.lnCatLevel = QtWidgets.QLineEdit(self.groupBox_5)
        self.lnCatLevel.setGeometry(466, 71, 46, 21)
        self.lnCatLevel.setText(str(ReSCUNetDialog.concatenation_level))

        # ============================== GroupBox with advanced options for Callbacks ==================================

        self.groupBox_7 = QtWidgets.QGroupBox(NNet)
        self.groupBox_7.setObjectName("groupBox_7")
        self.groupBox_7.setGeometry(30, 365, 551, 150)

        self.cbEarlyStopper = QtWidgets.QCheckBox(self.groupBox_7)
        self.cbEarlyStopper.setObjectName("cbEarlyStopper")
        self.cbEarlyStopper.setGeometry(31, 30, 250, 24)
        self.cbEarlyStopper.setChecked(ReSCUNetDialog.early_stopper['active'])
        self.label_patience = QtWidgets.QLabel(self.groupBox_7)
        self.label_patience.setObjectName("label_patience")
        self.label_patience.setGeometry(250, 30, 141, 24)
        self.label_patience.setEnabled(ReSCUNetDialog.early_stopper['active'])
        self.lnESPatience = QtWidgets.QLineEdit(self.groupBox_7)
        self.lnESPatience.setObjectName("lnESPatience")
        self.lnESPatience.setGeometry(325, 30, 46, 21)
        self.lnESPatience.setText(str(ReSCUNetDialog.early_stopper['kwargs']['patience']))
        self.lnESPatience.setEnabled(ReSCUNetDialog.early_stopper['active'])
        self.cbEarlyStopper.stateChanged.connect(partial(self._connect_check, self.cbEarlyStopper,
                                                         [self.label_patience, self.lnESPatience],
                                                         True))

        self.cbLRScheduler = QtWidgets.QCheckBox(self.groupBox_7)
        self.cbLRScheduler.setObjectName("cbLRScheduler")
        self.cbLRScheduler.setGeometry(31, 60, 250, 24)
        self.cbLRScheduler.setChecked(ReSCUNetDialog.lr_scheduler['active'])
        self.label_lrpatience = QtWidgets.QLabel(self.groupBox_7)
        self.label_lrpatience.setObjectName("label_lrpatience")
        self.label_lrpatience.setGeometry(250, 60, 141, 24)
        self.label_lrpatience.setEnabled(ReSCUNetDialog.lr_scheduler['active'])
        self.lnLRPatience = QtWidgets.QLineEdit(self.groupBox_7)
        self.lnLRPatience.setObjectName("lnLRPatience")
        self.lnLRPatience.setGeometry(325, 60, 46, 21)
        self.lnLRPatience.setText(str(ReSCUNetDialog.lr_scheduler['kwargs']['patience']))
        self.lnLRPatience.setEnabled(ReSCUNetDialog.lr_scheduler['active'])
        self.cbLRScheduler.stateChanged.connect(partial(self._connect_check, self.cbLRScheduler,
                                                        [self.label_lrpatience, self.lnLRPatience],
                                                        True))

        self.cbModelCheckpoint = QtWidgets.QCheckBox(self.groupBox_7)
        self.cbModelCheckpoint.setObjectName("cbModelCheckpoint")
        self.cbModelCheckpoint.setGeometry(31, 90, 250, 24)
        self.cbModelCheckpoint.setChecked(ReSCUNetDialog.model_checkpoint['active'])
        self.cbCkptBestOnly = QtWidgets.QCheckBox(self.groupBox_7)
        self.cbCkptBestOnly.setObjectName("cbCkptBestOnly")
        self.cbCkptBestOnly.setGeometry(250, 90, 250, 24)
        self.cbCkptBestOnly.setMaximumWidth(175)
        self.cbCkptBestOnly.setChecked(ReSCUNetDialog.model_checkpoint['kwargs']['save_best_only'])
        self.cbModelCheckpoint.stateChanged.connect(partial(self._connect_check, self.cbModelCheckpoint,
                                                            [self.cbCkptBestOnly], True))

        self.cbLogging = QtWidgets.QCheckBox(self.groupBox_7)
        self.cbLogging.setObjectName("cbLogging")
        self.cbLogging.setGeometry(31, 120, 200, 24)
        self.cbLogging.setChecked(ReSCUNetDialog.logging['active'])

        # ======================================== Accept or Reject Buttons ============================================

        self.buttonBox = QtWidgets.QDialogButtonBox(NNet)
        self.buttonBox.setGeometry(240, 525, 341, 32)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel | QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")

        self.retranslateUi(NNet)
        self.buttonBox.accepted.connect(NNet.accept)
        self.buttonBox.rejected.connect(NNet.reject)
        QtCore.QMetaObject.connectSlotsByName(NNet)

    def retranslateUi(self, CNN):
        _translate = QtCore.QCoreApplication.translate
        CNN.setWindowTitle(_translate("NNet", "Train network"))
        self.groupBox_2.setTitle(_translate("NNet", "Project files"))
        self.label_ptf.setText(_translate("NNet", "training image folder"))
        self.label_sf.setText(_translate("NNet", "save folder"))
        self.btnSavePositive.setText(_translate("NNet", "..."))
        self.btnSaveSave.setText(_translate("NNet", "..."))
        self.cbGenerateNotebook.setText(_translate("NNet", "generate training notebook"))
        self.btnSaveNotebook.setText(_translate("NNet", "..."))
        self.groupBox_3.setTitle(_translate("NNet", "Network input size"))
        self.label_w.setText(_translate("NNet", "width"))
        self.label_h.setText(_translate("NNet", "height"))
        self.groupBox_5.setTitle(_translate("NNet", "Other parameters"))
        self.label_epochs.setText(_translate("NNet", "epochs"))
        self.label_lr.setText(_translate("NNet", "learning rate"))
        self.groupBox_6.setTitle(_translate("NNet", "Step size (testing)"))
        self.label_r.setText(_translate("NNet", "rows"))
        self.label_c.setText(_translate("NNet", "columns"))
        self.label_bs.setText(_translate("NNet", "batch size"))
        self.label_ew.setText(_translate("NNet", "erosion width"))
        self.label_val.setText(_translate("NNet", "validation split"))
        self.label_cat.setText(_translate("NNet", "concatenation level"))
        self.groupBox_7.setTitle(_translate("NNet", "Advanced settings"))
        self.cbEarlyStopper.setText(_translate("NNet", "early stopping"))
        self.label_patience.setText(_translate("NNet", "patience"))
        self.cbLRScheduler.setText(_translate("NNet", "learning rate scheduler"))
        self.label_lrpatience.setText(_translate("NNet", "patience"))
        self.cbModelCheckpoint.setText(_translate("NNet", "save model checkpoints"))
        self.cbCkptBestOnly.setText(_translate("NNet", "save best weights only"))
        self.cbLogging.setText(_translate("NNet", "log training progress"))

    def parameters(self) -> dict:
        # os.path.join(self.positive_training_folder, '') adds a slash at the end of the filename.
        ReSCUNetDialog.positive_training_folder = os.path.join(self.positive_training_folder_edit.text(), '')
        ReSCUNetDialog.save_folder = os.path.join(self.save_folder_edit.text(), '')

        # Because of the U-Net architecture, input layer dimensions must be divisible by 16 to avoid issues.
        input_size = (int(self.lnHeight.text()), int(self.lnWidth.text()))
        height = (input_size[0] // 16) * 16 if input_size[0] % 16 else input_size[0]
        width = (input_size[1] // 16) * 16 if input_size[1] % 16 else input_size[1]
        ReSCUNetDialog.train_image_size = (height, width)

        ReSCUNetDialog.step_sz = (int(self.lnRow.text()), int(self.lnColumn.text()))
        ReSCUNetDialog.epochs = int(self.lnEpochs.text())
        ReSCUNetDialog.learning_rate = float(self.lnEta.text())
        ReSCUNetDialog.mini_batch_size = int(self.lnBatchSz.text())
        ReSCUNetDialog.erosion_width = int(self.lnErosionWidth.text())
        ReSCUNetDialog.generate_notebook = self.cbGenerateNotebook.isChecked()
        ReSCUNetDialog.notebook_path = self.notebook_path_edit.text()
        ReSCUNetDialog.validation_split = float(self.lnValSplit.text())
        ReSCUNetDialog.concatenation_level = int(self.lnCatLevel.text())
        ReSCUNetDialog.early_stopper['active'] = self.cbEarlyStopper.isChecked()
        ReSCUNetDialog.early_stopper['kwargs']['patience'] = int(self.lnESPatience.text())
        ReSCUNetDialog.lr_scheduler['active'] = self.cbLRScheduler.isChecked()
        ReSCUNetDialog.lr_scheduler['kwargs']['patience'] = int(self.lnLRPatience.text())
        ReSCUNetDialog.model_checkpoint['active'] = self.cbModelCheckpoint.isChecked()
        ReSCUNetDialog.model_checkpoint['kwargs']['save_best_only'] = self.cbCkptBestOnly.isChecked()
        ReSCUNetDialog.logging['active'] = self.cbLogging.isChecked()

        attributes = inspect.getmembers(ReSCUNetDialog, lambda a: not (inspect.isroutine(a)))
        return dict([a for a in attributes if not (a[0].startswith('__') and a[0].endswith('__'))])

    def _open_positive_folder_dialog(self) -> bool:
        folder = RUtils.open_folder_dialog("Positive training folder", self.positive_training_folder)

        if folder == '' or folder is False or self.positive_training_folder_edit is None:
            return False

        self.positive_training_folder_edit.setText(folder)
        return True

    def _open_notebook_path_dialog(self) -> bool:
        start_folder = self.notebook_path_edit.text() if self.notebook_path_edit.text() != '' \
            else self.positive_training_folder_edit.text() if self.positive_training_folder_edit.text() != '' \
            else self.notebook_path
        folder = RUtils.open_folder_dialog(f"Results folder", start_folder)

        if folder == '' or folder is False:
            return False

        self.notebook_path_edit.setText(folder)
        return True

    def _open_save_folder_dialog(self) -> bool:
        folder = RUtils.open_folder_dialog("Save folder", self.save_folder)

        if folder == '' or folder is False or self.save_folder_edit is None:
            return False

        self.save_folder_edit.setText(folder)
        return True

    @staticmethod
    def _connect_check(leader: QtWidgets.QCheckBox, followers: list, enable_on_check: bool):
        for follower in followers:
            follower.setEnabled(leader.isChecked() == enable_on_check)



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import importlib
import os
from pathlib import Path
import shutil
import sys
from typing import Optional

from PyQt6 import QtWidgets

import pyjamas.pjscore as pjscore
from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.dialogs.dropdowndialog import DropDownDialog


class RCBPlugins(RCallback):
    def cbLoadPlugins(self) -> True:
        """
        Load plugins into the plugins menu. Searches for plugins under pjs.plugin_path.

        :return: True.
        """
        plugin_files = list(Path(self.pjs.plugin_path).rglob("*.py"))
        for f in plugin_files:
            self.install_plugin(str(f))

        return True

    def cbInstallPlugin(self, filename: Optional[str] = None):
        """
        Install plugin: Copy plugin to the plugins folder (pjs.plugin_path) and load it into the plugins menu.

        :param filename: path and file name of the plugin to be installed; a dialog opens if the file name is empty or None.
        :return: True if the plugin was installed, False if the install was cancelled.
        """

        # Get file name.
        if filename is None or filename is False or filename == '': # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            fname = QtWidgets.QFileDialog.getOpenFileName(None, 'Install plugin ...', self.pjs.cwd,
                                                          filter='PyJAMAS plugin (*' + pjscore.PyJAMAS.plugin_extension + ')')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
        if filename == '':
            return False

        self.install_plugin(filename)

        return True

    def install_plugin(self, plugin_path: str) -> bool:
        if not os.path.exists(plugin_path):
            return False

        path, name = os.path.split(plugin_path)
        modulename, ext = os.path.splitext(name)

        # Copy plugin to plugins folder (which should be in sys.path). Remove path from sys.path.
        dest_folder: str = os.path.join(self.pjs.plugin_path, modulename)

        if path != dest_folder:
            os.makedirs(dest_folder, exist_ok=True)
            shutil.copyfile(plugin_path, os.path.join(dest_folder, name))

        sys.path.append(dest_folder)

        PJSPlugin = importlib.import_module(modulename)

        # If plugin classes can only be called PJSPlugin:
        plugin = PJSPlugin.PJSPlugin(self.pjs)
        # If arbitrary class names are OK, import the inspect module above and:
        #thepluginclass = inspect.getmembers(sys.modules[PJSPlugin.__name__], inspect.isclass)[0][1]
        #plugin = thepluginclass(self.pjs)

        self.pjs.plugin_list.append(plugin)
        self.pjs.plugin_path_list.append(dest_folder)
        plugin.build_menu()

        return True

    def cbUninstallPlugin(self, plugin_name: str = None):
        """
        Uninstall plugin: delete the folder for a given plugin from the PyJAMAS plugins directory.

        :param plugin_name: name of the plugin to be deleted.
        :return: True if the plugin was uninstalled, False if the operation was cancelled.
        """
        if len(self.pjs.plugin_list) == 0:
            self.pjs.statusbar.showMessage("No plugins installed.")
            return False
        else:
            plugin_name_list = [this_plugin.name() for this_plugin in self.pjs.plugin_list]

        if plugin_name is None or plugin_name is False:
            dialog = DropDownDialog(plugin_name_list)
            dialog.show()
            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
            parameters = dialog.parameters()
            plugin_index = parameters["index"]
            if not continue_flag:
                return False

        else:
            if plugin_name not in plugin_name_list:
                self.pjs.statusbar.showMessage("Could not find specified plugin.")
                return False
            else:
                plugin_index = plugin_name_list.index(plugin_name)

        self.uninstall_plugin(plugin_index)

        return True

    def uninstall_plugin(self, plugin_index):
        shutil.rmtree(self.pjs.plugin_path_list[plugin_index])
        self.pjs.menuPlugins.removeAction(self.pjs.plugin_list[plugin_index].menu)
        self.pjs.plugin_list.pop(plugin_index)
        self.pjs.plugin_path_list.pop(plugin_index)
        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional, Tuple
from abc import abstractmethod

import os
import numpy
import gzip
import pickle
import matplotlib.pyplot as plt
import skimage.io as sio
import skimage.morphology as sm
import skimage.segmentation as ss
import skimage.transform as st

import keras
import keras.models as km
import keras.optimizers as ko
import keras.utils as ku
import keras.callbacks as kc

import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.tensorboard import SummaryWriter

import pyjamas.pjscore
from pyjamas.rimage.rimml.rimml import rimml
from pyjamas.rutils import RUtils


class rimneuralnet(rimml):
    OUTPUT_CLASSES: int = 2
    BATCH_SIZE: int = 1
    EPOCHS: int = 100
    LEARNING_RATE: float = 0.001
    STEP_SIZE: Tuple[int, int] = (rimml.TRAIN_IMAGE_SIZE[0]//8, rimml.TRAIN_IMAGE_SIZE[1]//8)
    EROSION_WIDTH: int = 0
    PATIENCE: int = 5
    SAVE_BEST_ONLY_FLAG: bool = True
    SAVE_CHECKPOINTS_FLAG: bool = True
    EARLY_STOPPER_FLAG: bool = True
    LR_SCHEDULER_FLAG: bool = True
    LOG_FLAG: bool = True

    VALIDATION_SPLIT: float = 0.1
    EARLY_STOPPER: dict = {'active': EARLY_STOPPER_FLAG, 'kwargs': {'patience': PATIENCE}}
    LR_SCHEDULER: dict = {'active': LR_SCHEDULER_FLAG, 'kwargs': {'patience': PATIENCE}}  # ReduceLROnPlateau callback
    MODEL_CHECKPOINT: dict = {'active': SAVE_CHECKPOINTS_FLAG, 'kwargs': {'save_best_only': SAVE_BEST_ONLY_FLAG}}
    LOGGING: dict = {'active': LOG_FLAG, 'kwargs': {}}

    def __init__(self, parameters: Optional[dict] = None):
        super().__init__(parameters)

        self.positive_training_folder: str = parameters.get('positive_training_folder')
        self.save_folder: str = parameters.get('save_folder')

        # Size of training images (rows, columns).
        self.train_image_size: Tuple[int, int] = parameters.get('train_image_size', rimneuralnet.TRAIN_IMAGE_SIZE)  # (row, col)
        self.step_sz: Tuple[int, int] = parameters.get('step_sz', rimneuralnet.STEP_SIZE)

        self.scaler: int = parameters.get('scaler', 1)  # max pixel value of the training set.

        self.X_train: numpy.ndarray = None
        self.Y_train: numpy.ndarray = None
        self.W_train: numpy.ndarray = None

        self.save_dict: dict = {}

        self.object_array: numpy.ndarray = None
        self.prob_array: numpy.ndarray = None

        self.output_classes: int = parameters.get('output_classes', rimneuralnet.OUTPUT_CLASSES)
        self.learning_rate: float = parameters.get('learning_rate', rimneuralnet.LEARNING_RATE)
        self.input_size: Tuple[int, int] = parameters.get('train_image_size', rimneuralnet.TRAIN_IMAGE_SIZE)
        self.epochs: int = parameters.get('epochs', rimneuralnet.EPOCHS)
        self.mini_batch_size: int = parameters.get('mini_batch_size', rimneuralnet.BATCH_SIZE)
        self.erosion_width: int = parameters.get('erosion_width', rimneuralnet.EROSION_WIDTH)
        self.step_sz = parameters.get('step_sz', rimneuralnet.STEP_SIZE)
        self.validation_split = parameters.get('validation_split', rimneuralnet.VALIDATION_SPLIT)

        self.early_stopper = parameters.get('early_stopper', rimneuralnet.EARLY_STOPPER)
        self.lr_scheduler = parameters.get('lr_scheduler', rimneuralnet.LR_SCHEDULER)
        self.model_checkpoint = parameters.get('model_checkpoint', rimneuralnet.MODEL_CHECKPOINT)
        self.logging = parameters.get('logging', rimneuralnet.LOGGING)

        classifier_representation = parameters.get('classifier')
        if type(classifier_representation) is km.Model:
            self.classifier = classifier_representation
        else:
            if len(self.input_size) == 2:
                self.input_size = self.input_size + (1,)

        self.classifier = self.build_network(self.input_size, self.output_classes)
        adam = ko.Adam(learning_rate=self.learning_rate)
        self.classifier.compile(adam, loss=self.pixelwise_loss)
        if type(classifier_representation) is list:
            self.classifier.set_weights(classifier_representation)

        self.device = None
        if torch.mps.is_available():
            self.device = torch.device("mps")
        elif torch.cuda.is_available():
            self.device = torch.device("cuda")
        else:
            self.device = torch.device("cpu")

    @abstractmethod
    def build_network(self, input_shape: Tuple, n_classes: int) -> km.Model:
        pass

    @staticmethod
    def pixelwise_loss(target, output):
        """
        A custom function defined to simply sum the pixelwise loss.
        This function doesn't compute the crossentropy loss, since that is made a
        part of the model's computational graph itself.
        Parameters
        ----------
        target : keras.tensor
            A tensor corresponding to the true labels of an image.
        output : keras.tensor
            Model output
        Returns
        -------
        keras.tensor
            A tensor holding the aggregated loss.
        """
        return - keras.ops.sum(target * output,
                               len(output.shape) - 1)

    """def fit(self):
        self._process_training_data()
        inputs = self._get_training_inputs()
        callbacks = self._build_callbacks()

        self.classifier.fit(inputs, self.Y_train, batch_size=self.mini_batch_size, epochs=self.epochs,
                            validation_split=self.validation_split, callbacks=callbacks if callbacks else None)
        return True"""

    def fit(self):
        self._process_training_data()
        inputs = self._get_training_inputs()
        callbacks = self._build_callbacks()

        logs = {}
        writer = None

        if self.logging['active']:
            log_dir = os.path.join(self.save_folder, 'logs')
            os.makedirs(log_dir, exist_ok=True)
            writer = SummaryWriter(log_dir)

        # Convert to torch tensors
        tensor_inputs = [torch.tensor(inp, dtype=torch.float32).to(self.device) for inp in inputs]
        Y_train = torch.tensor(self.Y_train, dtype=torch.float32).to(self.device)

        # Split validation data
        n_samples = len(tensor_inputs[0])
        val_size = int(n_samples * self.validation_split)
        train_size = n_samples - val_size

        # Split data
        train_inputs = [inp[:train_size] for inp in tensor_inputs]
        val_inputs = [inp[train_size:] for inp in tensor_inputs]
        train_Y, val_Y = Y_train[:train_size], Y_train[train_size:]

        # Create DataLoaders
        train_dataset = TensorDataset(*train_inputs, train_Y)
        train_loader = DataLoader(train_dataset, batch_size=self.mini_batch_size, shuffle=True)

        if self.validation_split > 0 and val_size > 0:
            val_dataset = TensorDataset(*val_inputs, val_Y)
            val_loader = DataLoader(val_dataset, batch_size=self.mini_batch_size, shuffle=False)

        # Setup optimizer (keep the same learning rate)
        optimizer = torch.optim.Adam(self.classifier.parameters(), lr=self.learning_rate)

        # Initialize callback state
        best_loss = float('inf')
        patience_counter = 0
        lr_patience_counter = 0

        avg_val_loss = 0.
        avg_train_loss = 0.

        # Training loop
        for epoch in range(self.epochs):
            # Training phase
            self.classifier.train()
            train_loss = 0.0
            n_train_batches = 0

            for batch_data in train_loader:
                # Unpack: last element is Y, everything before is model inputs
                *batch_inputs, batch_Y = batch_data

                optimizer.zero_grad()

                # Forward pass - note your model expects [X, W] as inputs
                outputs = self.classifier(batch_inputs)

                # Your custom loss (already includes the weight map in the output)
                loss = -torch.sum(batch_Y * outputs, dim=-1).mean()

                loss.backward()
                optimizer.step()

                train_loss += loss.item()
                n_train_batches += 1

            avg_train_loss = train_loss / n_train_batches

            # Validation phase
            if self.validation_split > 0 and val_size > 0:
                self.classifier.eval()
                val_loss = 0.0
                n_val_batches = 0

                with torch.no_grad():
                    for batch_data in val_loader:
                        *batch_inputs, batch_Y = batch_data

                        outputs = self.classifier(batch_inputs)
                        loss = -torch.sum(batch_Y * outputs, dim=-1).mean()
                        val_loss += loss.item()
                        n_val_batches += 1

                avg_val_loss = val_loss / n_val_batches
                monitor_loss = avg_val_loss
                print(f"Epoch {epoch + 1}/{self.epochs} - loss: {avg_train_loss:.8f} - val_loss: {avg_val_loss:.8f}")
            else:
                monitor_loss = avg_train_loss
                print(f"Epoch {epoch + 1}/{self.epochs} - loss: {avg_train_loss:.8f}")

            logs = {
                'loss': avg_train_loss,
                'val_loss': avg_val_loss if self.validation_split > 0 else None
            }

            # Log to TensorBoard (if active)
            if writer:
                writer.add_scalar('Loss/train', avg_train_loss, epoch)
                if self.validation_split > 0:
                    writer.add_scalar('Loss/validation', avg_val_loss, epoch)
                writer.add_scalar('Learning_rate', optimizer.param_groups[0]['lr'], epoch)

            # Handle Early Stopping
            if self.early_stopper['active']:
                if monitor_loss < best_loss:
                    best_loss = monitor_loss
                    patience_counter = 0
                    # Save best weights
                    if self.model_checkpoint['active'] and self.model_checkpoint['kwargs'].get('save_best_only', True):
                        for callback in callbacks:
                            if isinstance(callback, SaveCallback):
                                callback.on_epoch_end(epoch, logs)
                else:
                    patience_counter += 1
                    if patience_counter >= self.early_stopper['kwargs']['patience']:
                        print(f"Early stopping triggered at epoch {epoch + 1}")
                        break

            # Handle Learning Rate Reduction
            if self.lr_scheduler['active']:
                if monitor_loss < best_loss:
                    lr_patience_counter = 0
                else:
                    lr_patience_counter += 1
                    if lr_patience_counter >= self.lr_scheduler['kwargs']['patience']:
                        old_lr = optimizer.param_groups[0]['lr']
                        new_lr = old_lr * 0.1  # Default factor for ReduceLROnPlateau
                        optimizer.param_groups[0]['lr'] = new_lr
                        print(f"Reducing learning rate to {new_lr}")
                        lr_patience_counter = 0

            # Call other callbacks (logging, model checkpoint)
            for callback in callbacks:
                if not isinstance(callback, SaveCallback) or not self.model_checkpoint['kwargs'].get('save_best_only',
                                                                                                     True):
                    if hasattr(callback, 'on_epoch_end'):
                        callback.on_epoch_end(epoch, logs)

        # Call on_train_end for callbacks
        for callback in callbacks:
            if hasattr(callback, 'on_train_end'):
                callback.on_train_end(logs)

        if writer:
            writer.close()

        return True

    def _process_training_data(self) -> bool:
        """
        Loads and scales training data and calculates weight maps.
        :return:
        """
        train_ids = next(os.walk(self.positive_training_folder))[1]

        # Get and resize train images and masks
        self.X_train = numpy.zeros((len(train_ids), self.train_image_size[0], self.train_image_size[1], 1), dtype=numpy.uint16)
        self.Y_train = numpy.zeros((len(train_ids), self.train_image_size[0], self.train_image_size[1], 1), dtype=bool)
        self.W_train = numpy.zeros((len(train_ids), self.train_image_size[0], self.train_image_size[1], 1), dtype=float)
        print('Resizing train images and masks and calculating weight maps ... ')

        for n, id_ in enumerate(train_ids):
            path = os.path.join(self.positive_training_folder, id_)
            im_file = os.path.join(path, "image", os.listdir(os.path.join(path, "image"))[0])
            img = sio.imread(im_file)
            if img.ndim == 3:
                img = img[0, :, :]
            img = numpy.expand_dims(st.resize(img, (self.train_image_size[0], self.train_image_size[1]), order=3,
                                              mode='constant', preserve_range=True), axis=-1)
            self.X_train[n] = img
            msk_file = os.path.join(path, "mask", os.listdir(os.path.join(path, "mask"))[0])
            mask = numpy.zeros((self.train_image_size[0], self.train_image_size[1], 1), dtype=bool)
            mask_ = sio.imread(msk_file)
            mask_ = numpy.expand_dims(st.resize(mask_, (self.train_image_size[0], self.train_image_size[1]), order=0,
                                                mode='constant', preserve_range=True), axis=-1)
            mask = numpy.maximum(mask, mask_)
            weights = self.weight_map(mask)
            self.Y_train[n] = mask
            self.W_train[n, :, :, 0] = weights

        self.scaler = numpy.amax(self.X_train)
        self.X_train = self.X_train / self.scaler

        wmap = numpy.zeros((self.X_train.shape[0], self.train_image_size[0], self.train_image_size[1], 2),
                           dtype=numpy.float32)
        wmap[..., 0] = self.W_train.squeeze()
        wmap[..., 1] = wmap[..., 0]
        self.W_train = wmap

        self.Y_train = ku.to_categorical(self.Y_train)
        return True

    def weight_map(self, binmasks: numpy.ndarray, w0: float = 10., sigma: float = 5., show: bool = False):
        """Compute the weight map for a given mask, as described in Ronneberger et al.
        (https://arxiv.org/pdf/1505.04597.pdf)
        """

        labmasks = sm.label(binmasks)
        n_objs = numpy.amax(labmasks)

        if n_objs == 0:
            # No objects, return uniform weights
            return numpy.ones_like(binmasks, dtype=numpy.float32)

        nrows, ncols = labmasks.shape[:2]

        # Move to GPU
        labmasks_gpu = torch.from_numpy(labmasks).to(self.device)
        masks = torch.zeros((n_objs, nrows, ncols), device=self.device)
        distMap = torch.zeros((nrows * ncols, n_objs), device=self.device)

        # Create meshgrid on GPU
        X1_gpu = torch.arange(nrows, device=self.device).repeat_interleave(ncols)
        Y1_gpu = torch.arange(ncols, device=self.device).repeat(nrows)

        for i in range(n_objs):
            mask = torch.squeeze(labmasks_gpu == i + 1)

            mask_cpu = mask.cpu().numpy()
            bounds = ss.find_boundaries(mask_cpu, mode='inner')
            X2, Y2 = numpy.nonzero(bounds)

            if len(X2) == 0:
                # No boundary pixels, skip this object
                mask[i] = mask
                continue

            # Move boundary coordinates to GPU
            X2_gpu = torch.from_numpy(X2).to(self.device).float()
            Y2_gpu = torch.from_numpy(Y2).to(self.device).float()

            xSum = (X2_gpu.reshape(-1, 1) - X1_gpu.reshape(1, -1)) ** 2
            ySum = (Y2_gpu.reshape(-1, 1) - Y1_gpu.reshape(1, -1)) ** 2
            distMap[:, i] = torch.sqrt(xSum + ySum).min(dim=0)[0]
            masks[i] = mask

        ix = torch.arange(distMap.shape[0], device=self.device)
        if distMap.shape[1] == 1:
            d1 = distMap.ravel()
            border_loss_map = w0 * torch.exp((-1 * (d1) ** 2) / (2 * (sigma ** 2)))
        else:
            if distMap.shape[1] == 2:
                # Get indices of 2 smallest distances
                _, indices = torch.topk(distMap, k=2, dim=1, largest=False)
                d1_ix, d2_ix = indices[:, 0], indices[:, 1]
            else:
                _, indices = torch.topk(distMap, k=2, dim=1, largest=False)
                d1_ix, d2_ix = indices[:, 0], indices[:, 1]

            d1 = distMap[ix, d1_ix]
            d2 = distMap[ix, d2_ix]
            border_loss_map = w0 * torch.exp((-1 * (d1 + d2) ** 2) / (2 * (sigma ** 2)))
        xBLoss = torch.zeros((nrows, ncols), device=self.device)
        xBLoss[X1_gpu, Y1_gpu] = border_loss_map
        # class weight map
        loss = torch.zeros((nrows, ncols), device=self.device)
        masks_sum = masks.sum()
        w_1 = 1 - masks_sum / loss.numel()
        w_0 = 1 - w_1

        mask_sum_dim0 = masks.sum(dim=0)
        loss[mask_sum_dim0 == 1] = w_1
        loss[mask_sum_dim0 == 0] = w_0

        ZZ = xBLoss + loss

        ZZ_cpu = ZZ.cpu().numpy()
        # ZZ = resize(ZZ, outsize, preserve_range=True)
        if show:
            plt.imshow(ZZ_cpu)
            plt.colorbar()
            plt.axis('off')
        return ZZ_cpu

    def _get_training_inputs(self):
        return [self.X_train, self.W_train]

    def _build_callbacks(self):
        callbacks = []
        if self.early_stopper['active']:
            if self.validation_split > 0:
                monitor = 'val_loss'
            else:
                raise Warning("Using early stopping without a validation split is not recommended. "
                              "Setting early stopping to monitor train loss.")
                monitor = 'train_loss'
            callbacks.append(kc.EarlyStopping(monitor, restore_best_weights=True, **self.early_stopper['kwargs']))
        if self.lr_scheduler['active']:
            callbacks.append(kc.ReduceLROnPlateau(monitor='val_loss' if self.validation_split > 0 else 'train_loss',
                                                  **self.lr_scheduler['kwargs']))

        if self.model_checkpoint['active']:  # add saving last so that it occurs after all other callback operations
            callbacks.append(SaveCallback(self, monitor='val_loss' if self.validation_split > 0 else 'train_loss',
                                          **self.model_checkpoint['kwargs']))

        # Set model reference for all callbacks
        for callback in callbacks:
            callback.set_model(self.classifier)

        return callbacks

    def build_save_dict(self) -> bool:
        self.save_dict.update({
            'positive_training_folder': self.positive_training_folder,
            'train_image_size': self.train_image_size,
            'scaler': self.scaler,
            'epochs': self.epochs,
            'mini_batch_size': self.mini_batch_size,
            'learning_rate': float(self.classifier.optimizer.learning_rate),
            'classifier': self.classifier.get_weights(),
            'step_sz': self.step_sz,
            'erosion_width': self.erosion_width,
            'early_stopper': self.early_stopper,
            'lr_scheduler': self.lr_scheduler,
            'model_checkpoint': self.model_checkpoint,
            'logging': self.logging,
        })
        return True

    def save(self, filename: str) -> bool:
        self.build_save_dict()
        return RUtils.pickle_this(self.save_dict, RUtils.set_extension(filename, pyjamas.pjscore.PyJAMAS.classifier_extension))


class SaveCallback(kc.Callback):
    """This is quite slow, use the ModelCheckpoint callback and implement a "load model weights" in the IO menu?"""

    def __init__(self, neuralnet: rimneuralnet, save_best_only: bool = True, monitor: str = 'val_loss'):
        super().__init__()

        if neuralnet.save_folder is None or neuralnet.save_folder == "" or neuralnet.save_folder is False:
            self.save_folder = os.getcwd()
        else:
            self.save_folder = neuralnet.save_folder

        self.neuralnet = neuralnet
        self.save_best_only = save_best_only
        self.monitor = monitor
        if save_best_only:
            self.best_loss = numpy.inf

    def on_epoch_end(self, epoch, logs=None):
        if self.save_best_only and logs[self.monitor] < self.best_loss:
            self.neuralnet.save(os.path.join(self.save_folder, "ckpt_best.cfr"))
            self.best_loss = logs[self.monitor]
        elif not self.save_best_only:
            self.neuralnet.save(os.path.join(self.save_folder, f"ckpt_epoch{epoch:03d}.cfr"))
        return {"epoch": epoch, "logs": logs}

    def on_train_end(self, logs):
        if self.save_best_only:
            fh = gzip.open(os.path.join(self.save_folder, "ckpt_best.cfr"), "rb")
            best_cfr = pickle.load(fh)
            fh.close()
            self.neuralnet.classifier.set_weights(best_cfr["classifier"])
        self.neuralnet.save(os.path.join(self.save_folder, "ckpt_final.cfr"))



from functools import partial
from typing import Optional
from PyQt6 import QtCore, QtWidgets, QtGui
from functools import partial
import time


class ConfigureKeyBindingsDialog(QtWidgets.QWidget):
    key_configs: dict = None
    searching: bool = False
    bttn: QtWidgets.QPushButton = None
    menu: str = None
    option: str = None
    dialog: QtWidgets.QDialog = None

    def __init__(self):
        super().__init__()

    def setupUi(self, Dialog: QtWidgets.QDialog, key_configs: dict):
        self.key_configs = key_configs
        self.dialog = Dialog

        self.gScene = QtWidgets.QGraphicsScene(self.dialog)
        self.gScene.setSceneRect(0, 0, 1183, 761)
        self.gScene.setObjectName('gScene')

        self.gView = QtWidgets.QGraphicsView(self.gScene)
        self.gView.setObjectName('gView')
        self.gView.installEventFilter(self)

        self.dialog.setObjectName("Dialog")
        self.dialog.resize(600, 500)
        self.buttonsOkCancel = QtWidgets.QDialogButtonBox(self.dialog)
        self.buttonsOkCancel.setGeometry(QtCore.QRect(0, 460, 560, 32))
        self.buttonsOkCancel.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonsOkCancel.setStandardButtons(
            QtWidgets.QDialogButtonBox.StandardButton.Cancel | QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonsOkCancel.setObjectName("buttonsOkCancel")

        self.TabBox = QtWidgets.QTabWidget(self.dialog)
        self.TabBox.setGeometry(QtCore.QRect(20, 20, 560, 430))
        self.TabBox.setObjectName("tab_box")

        self.menu_groupboxes = {}
        self.menu_scrollareas = {}
        self.option_labels = {}
        self.option_pbs = {}

        for this_menu in self.key_configs.keys():
            menu_configs = self.key_configs[this_menu]
            self.menu_groupboxes[this_menu] = QtWidgets.QGroupBox()
            self.menu_groupboxes[this_menu].setGeometry(QtCore.QRect(0, 0, 560, 430))
            self.menu_groupboxes[this_menu].setObjectName(f"menu_groupbox_{this_menu}")

            column_form = QtWidgets.QFormLayout()
            column_form.setFormAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
            column_form.setFormAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
            column_form.setObjectName(f"{this_menu}_columns")

            self.option_labels[this_menu] = {}
            self.option_pbs[this_menu] = {}
            for menu_option in menu_configs.keys():
                self.option_labels[this_menu][menu_option] = QtWidgets.QLabel(menu_option)
                self.option_pbs[this_menu][menu_option] = QtWidgets.QPushButton()
                self.option_pbs[this_menu][menu_option].setFixedSize(QtCore.QSize(100, 30))
                self.option_pbs[this_menu][menu_option].clicked.connect(
                    partial(self.button_clicked, self.option_pbs[this_menu][menu_option], this_menu, menu_option))
                column_form.addRow(self.option_pbs[this_menu][menu_option],
                                   self.option_labels[this_menu][menu_option])

            self.menu_groupboxes[this_menu].setLayout(column_form)
            self.menu_scrollareas[this_menu] = QtWidgets.QScrollArea()
            self.menu_scrollareas[this_menu].setWidget(self.menu_groupboxes[this_menu])
            self.menu_scrollareas[this_menu].setWidgetResizable(True)
            self.menu_scrollareas[this_menu].setFixedHeight(400)

            self.TabBox.addTab(self.menu_scrollareas[this_menu], this_menu)

        self.retranslateUi()
        self.buttonsOkCancel.accepted.connect(self.dialog.accept)
        self.buttonsOkCancel.rejected.connect(self.dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(self.dialog)

    def retranslateUi(self):
        _translate = QtCore.QCoreApplication.translate
        self.dialog.setWindowTitle(_translate("Dialog", "Configure key bindings"))

        for this_menu in self.key_configs.keys():
            for menu_option, option_shortcut in zip(self.key_configs[this_menu].keys(),
                                                    self.key_configs[this_menu].values()):
                self.option_labels[this_menu][menu_option].setText(_translate("Dialog", menu_option))
                self.option_pbs[this_menu][menu_option].setText(_translate("Dialog", option_shortcut))

    def parameters(self) -> dict:
        for this_menu in self.key_configs.keys():
            for menu_option in self.key_configs[this_menu].keys():
                self.key_configs[this_menu][menu_option] = self.option_pbs[this_menu][menu_option].text()

        ConfigureKeyBindingsDialog.key_configs: dict = self.key_configs
        return ConfigureKeyBindingsDialog.key_configs

    def button_clicked(self, bttn: QtWidgets.QPushButton, menu: str, option: str):
        _translate = QtCore.QCoreApplication.translate
        bttn.setText(_translate('Dialog', '...'))
        self.gView.grabKeyboard()
        self.bttn = bttn
        self.menu = menu
        self.option = option
        self.searching = True

    def eventFilter(self, source, event: QtCore.QEvent):
        modifier_keys = [QtCore.Qt.Key.Key_Shift, QtCore.Qt.Key.Key_Alt,
                         QtCore.Qt.Key.Key_Control, QtCore.Qt.Key.Key_Meta]
        if type(source) == QtWidgets.QGraphicsView and self.searching:
            if event.type() == QtCore.QEvent.Type.KeyPress:
                if event.key() not in modifier_keys:
                    if event.key() == QtCore.Qt.Key.Key_Escape:
                        self.finish_search("")
                    else:
                        self.finish_search(QtGui.QKeySequence(event.keyCombination()).toString())
                        return True
        return False

    def finish_search(self, key_string: str):
        self.searching = False
        if self.bttn is not None:
            if key_string != "":  # If not empty string, replace matching shortcuts with empty strings
                for this_menu in self.key_configs.keys():
                    for this_option in self.key_configs[this_menu].keys():
                        if self.key_configs[this_menu][this_option] == key_string:
                            self.key_configs[this_menu][this_option] = ""
            self.key_configs[self.menu][self.option] = key_string
        self.retranslateUi()
